#include "OFFFile.h"

OFFFile::OFFFile() :
    disp_list(0)
{

}

OFFFile::~OFFFile()
{   

    Clear();

}

void OFFFile::Clear()
{

    m_vVertices.clear();
    m_vVertNorms.clear();
    m_vFaceNorms.clear();
    m_vFace_v.clear();
    m_vFace_n.clear();

    if (disp_list) {
        glDeleteLists(disp_list, 1);
        disp_list = 0;
    }

}

bool OFFFile::Read(const QString & filename)
{ 

    Clear();

    QFile file(filename);

    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "OFF File " << filename << " can't be opened for reading";
        return false;
    }

    QTextStream ifs(&file);

    if (ifs.readLine().left(1) != "O") {
        qDebug() << "OFF File " << filename << " has invalid format";
        return false;
    }

    int nVerts, nFaces;

    QStringList eachlist = ifs.readLine().split(" ");
    nVerts = eachlist[0].toInt();
    nFaces = eachlist[1].toInt();

    m_vVertices.resize(nVerts);
    m_vFace_v.resize(nFaces);

    for (int i=0; i<nVerts; ++i) {

        eachlist = ifs.readLine().split(" ");

        m_vVertices[i][0] = eachlist[0].toFloat();
        m_vVertices[i][1] = eachlist[1].toFloat();
        m_vVertices[i][2] = eachlist[2].toFloat();

    }   

    for (int i=0; i<nFaces; ++i) {

        eachlist = ifs.readLine().split(" ");

        int nFaceVerts = eachlist.first().toInt();

        eachlist.pop_front();

        m_vFace_v[i].resize(nFaceVerts);

        for (int j=0; j<nFaceVerts; ++j) {

            m_vFace_v[i][j] = eachlist[j].toInt();

        }

    }

    file.close();

    ComputeFaceNormals();

    return true;

}

bool OFFFile::Read_OBJ(const QString & filename)
{

    Clear();

    QFile file(filename);

    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "OFF File " << filename << " can't be opened for reading";
        return false;
    }

    QTextStream ifs(&file);

    QList <vec3> verts;
    QList <vec3> vert_norms;
    QList <QVector <int> > facevert;
    QList <QVector <int> > facenorm;

    while (!ifs.atEnd()) {

        QStringList eachline = ifs.readLine().split(" ");

        if (QString::compare(eachline.first(), "v") == 0) {

            vec3 v;
            v[0] = eachline[1].toDouble();
            v[1] = eachline[2].toDouble();
            v[2] = eachline[3].toDouble();

            verts.push_back(v);

        }
        else if (QString::compare(eachline.first(), "vn") == 0) {

            vec3 n;
            n[0] = eachline[1].toDouble();
            n[1] = eachline[2].toDouble();
            n[2] = eachline[3].toDouble();

            vert_norms.push_back(n);

        }
        else if (QString::compare(eachline.first(), "f") == 0) {

            eachline.pop_front();

            QVector <int> f_v(eachline.size());
            QVector <int> f_n(eachline.size());

            for (int i=0; i<eachline.size(); ++i) {

                QStringList each = eachline[i].split("/");
                f_v[i] = each[0].toInt() - 1;

                if (each.size() >= 2) {
                    f_n[i] = each[2].toInt() - 1;
                }

            }

            facevert.push_back(f_v);
            facenorm.push_back(f_n);

        }

    }

    file.close();

    m_vVertices = verts.toVector();
    m_vVertNorms = vert_norms.toVector();
    m_vFace_v = facevert.toVector();
    m_vFace_n = facenorm.toVector();

    ComputeFaceNormals();

    return true;

}

vec3 OFFFile::GetVert(const int i)
{
    return m_vVertices[i];
}

QVector <vec3> & OFFFile::GetVerts()
{
    return m_vVertices;
}

const QVector <vec3> & OFFFile::GetVerts() const
{
    return m_vVertices;
}

QVector <int> OFFFile::GetFace(const int i)
{
    return m_vFace_v[i];
}

QVector <QVector <int> > & OFFFile::GetFaces()
{
    return m_vFace_v;
}

const QVector <QVector <int> > & OFFFile::GetFaces() const
{
    return m_vFace_v;
}

QVector <vec3> & OFFFile::GetFaceNorms()
{
    return m_vFaceNorms;
}

const QVector <vec3> & OFFFile::GetFaceNorms() const
{
    return m_vFaceNorms;
}

vec3 OFFFile::GetFaceCentroid(const int face_index) const
{

    vec3 face_centroid = vec3(0, 0, 0);

    for (int i=0; i<m_vFace_v[face_index].size(); ++i) {
        face_centroid += m_vVertices[m_vFace_v[face_index][i]];
    }

    face_centroid *= 1.0f / float(m_vFace_v[face_index].size());

    return face_centroid;

}

void OFFFile::Normalize()
{

    vec3 minv(FLT_MAX, FLT_MAX, FLT_MAX);
    vec3 maxv(-FLT_MAX, -FLT_MAX, -FLT_MAX);

    for (int i=0; i<m_vVertices.size(); ++i) {

        vec3 & v = m_vVertices[i];

        for (int j=0; j<3; ++j) {
            minv[j] = qMin(minv[j], v[j]);
            maxv[j] = qMax(maxv[j], v[j]);
        }

    }

    vec3 vdiff = maxv - minv;
    vec3 cent = minv + vdiff * 0.5f;
    float diff = qMax(vdiff[0], qMax(vdiff[1], vdiff[2]));

    for (int i=0; i<m_vVertices.size(); ++i) {

        vec3 & v = m_vVertices[i];

        for (int j=0; j<3; ++j) {
            v[j] = (v[j] - cent[j]) / diff;
        }

    }

}

bool OFFFile::Write_OBJ(const QString & filename) const
{

    QFile file(filename);

    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qDebug() << "OBJ File " << filename << " can't be written";
        return false;
    }

    qDebug() << "Writing" << filename << "...";

    QTextStream ofs(&file);

    ofs << "#" << m_vVertices.size() << "\n";
    ofs << "#" << m_vFace_v.size() << "\n";

    for (int i=0; i<m_vVertices.size(); ++i) {

        ofs << "v " << m_vVertices[i][0] << " " << m_vVertices[i][1] << " " << m_vVertices[i][2] << "\n";

    }

    for (int i=0; i<m_vFace_v.size(); ++i) {

        ofs << "f";

        //for (int j=m_vFaces[i].size()-1; j>=0; --j) {
        for (int j=0; j<m_vFace_v[i].size(); ++j) {

            ofs << " " << (m_vFace_v[i][j] + 1);

        }

        ofs << "\n";

    }

    file.close();

    qDebug() << "Writing" << filename << "successful.";
    return true;

}

bool OFFFile::Write(const QString & filename) const
{

    QFile file(filename);

    if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
        qDebug() << "OFFFile " << filename << " can't be written";
        return false;
    }

    qDebug() << "Writing" << filename << "...";

    QTextStream ofs(&file);

    ofs << "OFF\n";

    ofs << m_vVertices.size() << " " << m_vFace_v.size() << " " << m_vFace_v.size() << "\n";

    for (int i=0; i<m_vVertices.size(); ++i) {

        ofs << m_vVertices[i][0] << " " << m_vVertices[i][1] << " " << m_vVertices[i][2] << "\n";
        //swap xy
        //ofs << m_vVertices[i][0] << " " << m_vVertices[i][2] << " " << m_vVertices[i][1] << "\n";

    }

    for (int i=0; i<m_vFace_v.size(); ++i) {
    //for (int i=m_vFaces.size()-1; i>=0; --i) {

        ofs << m_vFace_v[i].size();

        //for (int j=0; j<m_vFaces[i].size(); ++j) {
        for (int j=m_vFace_v[i].size()-1; j>=0; --j) {

            ofs << " " << m_vFace_v[i][j];

        }

        ofs << "\n";

    }

    file.close();

    qDebug() << "Writing" << filename << "successful.";
    return true;

}

void OFFFile::DrawGL()
{

    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    if (disp_list) {

        glCallList(disp_list);

    }
    else {

        disp_list = glGenLists(1);
        glNewList(disp_list, GL_COMPILE_AND_EXECUTE);

        glBegin(GL_TRIANGLES);

        for (int i=0; i<m_vFace_v.size(); ++i) {

            if (m_vVertNorms.empty()) {

                glNormal3f(m_vFaceNorms[i][0], m_vFaceNorms[i][1], m_vFaceNorms[i][2]);

                //draw all triangles to represent face
                for (int k=2; k<m_vFace_v[i].size(); ++k) {

                    const vec3 & v1 = m_vVertices[m_vFace_v[i][0]];
                    const vec3 & v2 = m_vVertices[m_vFace_v[i][k-1]];
                    const vec3 & v3 = m_vVertices[m_vFace_v[i][k]];

                    glVertex3f(v1[0], v1[1], v1[2]);
                    glVertex3f(v2[0], v2[1], v2[2]);
                    glVertex3f(v3[0], v3[1], v3[2]);

                }

            }
            else {

                //draw all triangles to represent face
                for (int k=2; k<m_vFace_v[i].size(); ++k) {

                    const vec3 & n1 = m_vVertNorms[m_vFace_n[i][0]];
                    const vec3 & n2 = m_vVertNorms[m_vFace_n[i][k-1]];
                    const vec3 & n3 = m_vVertNorms[m_vFace_n[i][k]];

                    const vec3 & v1 = m_vVertices[m_vFace_v[i][0]];
                    const vec3 & v2 = m_vVertices[m_vFace_v[i][k-1]];
                    const vec3 & v3 = m_vVertices[m_vFace_v[i][k]];

                    glNormal3f(n1[0], n1[1], n1[2]);
                    glVertex3f(v1[0], v1[1], v1[2]);

                    glNormal3f(n2[0], n2[1], n2[2]);
                    glVertex3f(v2[0], v2[1], v2[2]);

                    glNormal3f(n3[0], n3[1], n3[2]);
                    glVertex3f(v3[0], v3[1], v3[2]);

                }

            }

        }

        glEnd();

        glEndList();

    }

    glDisable(GL_LIGHTING);

}

void OFFFile::DrawCurvatureGL(const QList <float> & curvs)
{

    glEnable(GL_COLOR_MATERIAL);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);

    if (disp_list) {

        glCallList(disp_list);

    }
    else {

        /*
        QList <float> curvs2;
        for (int i=0; i<curvs.size(); ++i) {
            curvs2.push_back(curvs[i]);
        }
        */
        QList <float> curvs2 = curvs;
        qSort(curvs2);

        float max_curv = curvs2[curvs2.size()*80/100]; //for viewdep curvature
        //float max_curv = curvs2[curvs2.size()*90/100] * 2.0f; //for curvature
        if (max_curv <= 0.0f) {
            max_curv = curvs2.last();
        }
        //qDebug() << "max_curv" << max_curv;

        disp_list = glGenLists(1);
        glNewList(disp_list, GL_COMPILE_AND_EXECUTE);

        glBegin(GL_TRIANGLES);

        for (int i=0; i<m_vFace_v.size(); ++i) {

            glNormal3f(m_vFaceNorms[i][0], m_vFaceNorms[i][1], m_vFaceNorms[i][2]);

            //draw all triangles to represent face
            for (int k=2; k<m_vFace_v[i].size(); ++k) {

                const vec3 & v1 = m_vVertices[m_vFace_v[i][0]];
                const vec3 & v2 = m_vVertices[m_vFace_v[i][k-1]];
                const vec3 & v3 = m_vVertices[m_vFace_v[i][k]];

                SetColorByCurvature(curvs[m_vFace_v[i][0]], max_curv);
                glVertex3f(v1[0], v1[1], v1[2]);

                SetColorByCurvature(curvs[m_vFace_v[i][k-1]], max_curv);
                glVertex3f(v2[0], v2[1], v2[2]);

                SetColorByCurvature(curvs[m_vFace_v[i][k]], max_curv);
                glVertex3f(v3[0], v3[1], v3[2]);

            }

        }

        glEnd();

        glEndList();

    }

    glDisable(GL_LIGHTING);

}

void OFFFile::ComputeFaceNormals()
{

    // Process face normals
    m_vFaceNorms.clear();
    m_vFaceNorms.resize(m_vFace_v.size());

    for(int nFace=0; nFace<m_vFace_v.size(); ++nFace)
    {
        int vertIndex;
        //float triangle[3][3];
        vec3 v[3];

        for(int i = 0; i < m_vFace_v[nFace].size(); ++i)
        {
            // Get the vertex index for each point of the face
            vertIndex = m_vFace_v[nFace][i];
            v[i] = m_vVertices[vertIndex];

        }

        vec3 edge1, edge2;
        edge1 = v[0] - v[1];
        edge2 = v[2] - v[1];

        vec3 eachNorm = edge2 ^ edge1;
        eachNorm.normalize();

        m_vFaceNorms[nFace] = eachNorm;

    }

}

void OFFFile::FlipFaceNormals()
{

    for (int i=0; i<m_vFaceNorms.size(); ++i) {
        m_vFaceNorms[i] = -m_vFaceNorms[i] * (-1.0f);
    }

}

void OFFFile::AddFace(const vec3 & v1, const vec3 & v2, const vec3 & v3, const vec3 & v4)
{

    //add face with indexes of verts to add
    int nVerts = m_vVertices.size();

    QVector <int> face;

    face.push_back(nVerts);
    face.push_back(nVerts+1);
    face.push_back(nVerts+2);
    face.push_back(nVerts+3);

    m_vFace_v.push_back(face);

    //now add the verts
    m_vVertices.push_back(v1);
    m_vVertices.push_back(v2);
    m_vVertices.push_back(v3);
    m_vVertices.push_back(v4);

}

void OFFFile::AddFace(const vec3 & v1, const vec3 & v2, const vec3 & v3)
{

    //add face with indexes of verts to add
    int nVerts = m_vVertices.size();

    QVector <int> face;

    face.push_back(nVerts);
    face.push_back(nVerts+1);
    face.push_back(nVerts+2);

    m_vFace_v.push_back(face);

    //now add the verts
    m_vVertices.push_back(v1);
    m_vVertices.push_back(v2);
    m_vVertices.push_back(v3);

}


void OFFFile::AddCube(const vec3 & v, const float size)
{

    vec3 v1 = v + vec3(0, 0, 0);
    vec3 v2 = v + vec3(size, 0, 0);
    vec3 v3 = v + vec3(size, 0, size);
    vec3 v4 = v + vec3(0, 0, size);
    vec3 v5 = v + vec3(0, size, 0);
    vec3 v6 = v + vec3(size, size, 0);
    vec3 v7 = v + vec3(size, size, size);
    vec3 v8 = v + vec3(0, size, size);

    int n = m_vVertices.size();

    m_vVertices.push_back(v1);
    m_vVertices.push_back(v2);
    m_vVertices.push_back(v3);
    m_vVertices.push_back(v4);
    m_vVertices.push_back(v5);
    m_vVertices.push_back(v6);
    m_vVertices.push_back(v7);
    m_vVertices.push_back(v8);

    QVector <int> bottom_face;
    QVector <int> right_face;
    QVector <int> back_face;
    QVector <int> left_face;
    QVector <int> front_face;
    QVector <int> top_face;

    //bottom face
    bottom_face.push_back(n);
    bottom_face.push_back(n+3);
    bottom_face.push_back(n+2);
    bottom_face.push_back(n+1);

    //right face
    right_face.push_back(n+1);
    right_face.push_back(n+2);
    right_face.push_back(n+6);
    right_face.push_back(n+5);

    //back face
    back_face.push_back(n+2);
    back_face.push_back(n+3);
    back_face.push_back(n+7);
    back_face.push_back(n+6);

    //left face
    left_face.push_back(n);
    left_face.push_back(n+4);
    left_face.push_back(n+7);
    left_face.push_back(n+3);

    //front face
    front_face.push_back(n);
    front_face.push_back(n+1);
    front_face.push_back(n+5);
    front_face.push_back(n+4);

    //top face
    top_face.push_back(n+4);
    top_face.push_back(n+5);
    top_face.push_back(n+6);
    top_face.push_back(n+7);

    m_vFace_v.push_back(bottom_face);
    m_vFace_v.push_back(right_face);
    m_vFace_v.push_back(back_face);
    m_vFace_v.push_back(left_face);
    m_vFace_v.push_back(front_face);
    m_vFace_v.push_back(top_face);

}

void OFFFile::AddDisc(const vec3 & v, const vec3 & n, const float radius)
{

    //set vector basis so disc lies ortho to n
    TNBFrame frame;
    frame.SetN(n);

    float incr = 0.05f;

    QList <float> angles;
    for (float f=0.0f; f<ALG_M2_PI-incr; f+=incr) {
        angles.push_back(f);
    }
    angles.push_back(angles.first());

    for (int i=0; i<angles.size()-1; ++i) {

        float x1 = cosf(angles[i]);
        float y1 = sinf(angles[i]);

        float x2 = cosf(angles[i+1]);
        float y2 = sinf(angles[i+1]);

        vec3 v1 = v + (frame.B() * x1 * radius) + (frame.T() * y1 * radius);
        vec3 v2 = v + (frame.B() * x2 * radius) + (frame.T() * y2 * radius);

        vec3 shift = frame.N() * radius * 0.2f;

        //add verts for tri
        m_vVertices.push_back(v + shift);
        m_vVertices.push_back(v2 + shift);
        m_vVertices.push_back(v1 + shift);

        m_vVertices.push_back(v);
        m_vVertices.push_back(v2);
        m_vVertices.push_back(v1);

        int n = m_vVertices.size();

        QVector <int> top_face;

        top_face.push_back(n-6);
        top_face.push_back(n-5);
        top_face.push_back(n-4);

        QVector <int> bottom_face;

        bottom_face.push_back(n-3);
        bottom_face.push_back(n-1);
        bottom_face.push_back(n-2);

        QVector <int> side_face;

        side_face.push_back(n-2);
        side_face.push_back(n-1);
        side_face.push_back(n-4);
        side_face.push_back(n-5);

        m_vFace_v.push_back(top_face);
        m_vFace_v.push_back(bottom_face);
        m_vFace_v.push_back(side_face);

    }

}

float OFFFile::ComputeSurfaceArea(const bool special_case)
{

    float surface_area = 0.0f;

    for (int i=0; i<m_vFace_v.size(); ++i) {

        if (special_case && (i % 5) >= 2) {
            continue;
        }

       //for every face we fan the triangles on the first vertex in the face list
        for (int j=2; j<m_vFace_v[i].size(); ++j) {

            int ind1 = m_vFace_v[i][0];
            int ind2 = m_vFace_v[i][j-1];
            int ind3 = m_vFace_v[i][j];

            vec3 & v1 = m_vVertices[ind1];
            vec3 & v2 = m_vVertices[ind2];
            vec3 & v3 = m_vVertices[ind3];

            //area of triangle is half of parallelogram made by two vectors,
            //and parallelogram area given by magnitude of the cross product
            surface_area += ((v2 - v1) ^ (v3 - v1)).length() * 0.5f;

        }

    }

    return surface_area;

}

int OFFFile::GetClosestVertexIndex(const vec3 & p) const
{

    double min_dist = FLT_MAX;
    int min_index = 0;

    for (int i=0; i<m_vVertices.size(); ++i) {

        const double each_dist = (p - m_vVertices[i]).length();

        if (each_dist < min_dist) {
            min_dist = each_dist;
            min_index = i;
        }

    }

    return min_index;

}

void OFFFile::ApplyModelviewMatrix(const float mview [])
{

    for (int i=0; i<m_vVertices.size(); ++i) {

        m_vVertices[i] = ModelviewMultiply(mview, m_vVertices[i]);

    }

    ComputeFaceNormals();

}

void OFFFile::SetColorByCurvature(const vec2 curv, const float max_curv)
{

    GLubyte r, g, b;

    int val = int((fabsf(curv[0]) / max_curv) * 255.0f);
    val = qMax(val, 0);
    val = qMin(val, 255);

    if (curv[0] > 0.0) {

        r = 255;
        g = 255 - val;
        b = 255 - val;

    }
    else {

        r = 255 - val;
        g = 255 - val;
        b = 255;

    }

    glColor3ub(r, g, b);

}
