#include "algebra3.h"

QDebug operator<< (QDebug dbg, const vec2 & v)
{
    dbg << "vec2(" << v[0] << ", " << v[1] << ")";
    return dbg;
}

QDebug operator<< (QDebug dbg, const vec3 & v)
{
    dbg << "vec3(" << v[0] << ", " << v[1] << ", " << v[2] << ")";
    return dbg;
}

QDebug operator<< (QDebug dbg, const vec4 & v)
{
    dbg << "vec4(" << v[0] << ", " << v[1] << ", " << v[2] << "," << v[3] << ")";
    return dbg;
}


void NearestMajorAxis(const vec3 & v, vec3 & v_axis, float & v_angle)
{

    vec3 axes[6];
    axes[0] = vec3(1, 0, 0);
    axes[1] = vec3(0, 1, 0);
    axes[2] = vec3(0, 0, 1);
    axes[3] = vec3(-1, 0, 0);
    axes[4] = vec3(0, -1, 0);
    axes[5] = vec3(0, 0, -1);

    v_angle = FLT_MAX;

    for (int i=0; i<6; ++i) {

        float each_angle = AngleBetweenDeg(v, axes[i]);

        if (each_angle < v_angle) {

            v_angle = each_angle;
            v_axis = axes[i];

        }

    }

}

float AxisAngleDegrees(const vec3 & v1, const vec3 & v2)
{

    vec3 v1norm = v1;
    vec3 v2norm = v2;

    v1norm.normalize();
    v2norm.normalize();

    if (v1norm * v2norm < 0.0) {
        return fabsf(acosf(v1norm * (-v2norm))) * RAD_TO_DEG;
    }
    else {
        return fabsf(acosf(v1norm * v2norm)) * RAD_TO_DEG;
    }

}

vec3 RemoveFromVector(const vec3 & v, vec3 & removing_v)
{

    return v - removing_v * (v * removing_v);

}

vec3 LinearInterpolate(const vec3 & v1, const vec3 & v2, int t, int max_t)
{

    float interp = float(t) / float(max_t);
    return v1 * (1.0f - interp) + v2 * interp;

}

vec2 LinearInterpolate(const vec2 & v1, const vec2 & v2, int t, int max_t)
{

    float interp = float(t) / float(max_t);
    return v1 * (1.0f - interp) + v2 * interp;

}

vec3 CubicHermiteInterpolate(const vec3 & p0, const vec3 & t0, const vec3 & p1, const vec3 & t1, const float t)
{

    float t2 = t*t;
    float t3 = t2*t;

    return p0 * (2*t3 - 3*t2 + 1) + t0 * (t3 -2*t2 + t) + p1 * (-2*t3 + 3*t2) + (t3 - t2) * t1;

}

bool LineSegmentsIntersect2D(const vec2 & p1, const vec2 & p2, const vec2 & p3, const vec2 & p4, vec2 & int_pt, float & u1, float & u2)
{

    float u_denom = (p4[1] - p3[1]) * (p2[0] - p1[0]) - (p4[0] - p3[0]) * (p2[1] - p1[1]);

    if (fabs(u_denom) < 0.001f) {
        return false;
    }

    float ua_num = (p4[0] - p3[0]) * (p1[1] - p3[1]) - (p4[1] - p3[1]) * (p1[0] - p3[0]);
    float ub_num = (p2[0] - p1[0]) * (p1[1] - p3[1]) - (p2[1] - p1[1]) * (p1[0] - p3[0]);

    u1 = ua_num / u_denom;
    u2 = ub_num / u_denom;

    int_pt = p1 + (p2 - p1) * u1;

    return (u1 >= 0.0f && u1 <= 1.0f && u2 >= 0.0f && u2 <= 1.0f);

}

bool LinesIntersect2D(const vec2 & p1, const vec2 & p2, const vec2 & p3, const vec2 & p4, vec2 & int_pt, float & u1, float & u2)
{

    float u_denom = (p4[1] - p3[1]) * (p2[0] - p1[0]) - (p4[0] - p3[0]) * (p2[1] - p1[1]);

    if (fabs(u_denom) < 0.00001f) {
        return false;
    }

    float ua_num = (p4[0] - p3[0]) * (p1[1] - p3[1]) - (p4[1] - p3[1]) * (p1[0] - p3[0]);
    float ub_num = (p2[0] - p1[0]) * (p1[1] - p3[1]) - (p2[1] - p1[1]) * (p1[0] - p3[0]);

    u1 = ua_num / u_denom;
    u2 = ub_num / u_denom;

    int_pt = p1 + (p2 - p1) * u1;

    return true;

}

vec3 GetOrthoVec(const vec3 & v)
{

    //preference to Y-up
    vec3 cross_vec;

    if (fabs(v * vec3(0, 1, 0)) > 0.9f) {

        cross_vec = vec3(0, 0, 1) ^ v;
        cross_vec.normalize();
        cross_vec = cross_vec ^ v;

    }
    else {

        cross_vec = vec3(0, 1, 0) ^ v;
        cross_vec.normalize();
        cross_vec = v ^ cross_vec;

    }

    cross_vec.normalize();

    return cross_vec;
}


bool GetRayTriIntersect(const vec3 & rayp, const vec3 & rayd, const vec3 & tri_p1, const vec3 & tri_p2, const vec3 & tri_p3, vec3 & ipt)
{

    vec3 e1 = tri_p2 - tri_p1;
    vec3 e2 = tri_p3 - tri_p1;

    // Find the cross product of edge2 and the ray direction
    vec3 s1 = rayd ^ e2;

    // Find the divisor, if its zero, return false as the triangle is degenerate
    float divisor = s1 * e1;
    if (divisor == 0.0f) {
        return false;
    }

    // A inverted divisor, as multipling is faster then division
    float invDivisor = 1.0f / divisor;

    // Calculate the first barycentic coordinate. Barycentic coordinates
    // are between 0.0 and 1.0
    vec3 distance = rayp - tri_p1;
    float barycCoord_1 = (distance * s1) * invDivisor;

    if (barycCoord_1 < 0.0f || barycCoord_1 > 1.0f) {
        return false;
    }

    // Calculate the second barycentic coordinate
    vec3 s2 = distance ^ e1;
    float barycCoord_2 = (rayd * s2) * invDivisor;
    if (barycCoord_2 < 0.0f || (barycCoord_1 + barycCoord_2) > 1.0f) {
        return false;
    }

    // After doing the barycentic coordinate test we know if the ray hits or
    // not. If we got this far the ray hits.
    // Calculate the distance to the intersection point
    float intdist = (e2 * s2) * invDivisor;
    ipt = rayp + rayd * intdist;
    //rayp.Debug("rayp");
    //rayd.Debug("rayd");
    //qDebug() << "intdist" << intdist;
    return true;

}

bool LineLineIntersect(const vec3 & p1, const vec3 & p2, const vec3 & p3, const vec3 & p4, vec3 & pa, vec3 & pb, double & mua, double & mub)
{

    //code adapted from:
    //http://paulbourke.net/geometry/lineline3d/

    vec3 p13, p43, p21;
    double d1343, d4321, d1321, d4343, d2121;
    double numer, denom;

    double epsilon = 0.0001;

    p13 = p1 - p3;
    p43 = p4 - p3;
    if (p43.length() < epsilon) {
        return false;
    }

    p21 = p2 - p1;
    if (p21.length() < epsilon) {
        return false;
    }

    d1343 = p13 * p43;
    d4321 = p43 * p21;
    d1321 = p13 * p21;
    d4343 = p43 * p43;
    d2121 = p21 * p21;

    denom = d2121 * d4343 - d4321 * d4321;

    if (abs(denom) < epsilon) {
        return false;
    }
    numer = d1343 * d4321 - d1321 * d4343;

    mua = numer / denom;
    mub = (d1343 + d4321 * (mua)) / d4343;

    pa = p1 + p21 * mua;
    pb = p3 + p43 * mub;

    return true;

}

float AngleBetweenRad(const vec3 & v1, const vec3 & v2)
{

    vec3 v1norm = v1;
    vec3 v2norm = v2;

    v1norm.normalize();
    v2norm.normalize();

    return acosf(v1norm * v2norm);

}

float AngleBetweenDeg(const vec3 & v1, const vec3 & v2)
{

    return AngleBetweenRad(v1, v2) * RAD_TO_DEG;

}

vec3 ModelviewMultiply(const float mview[], const vec3 & p)
{

    return vec3(mview[0] * p[0] + mview[4] * p[1] + mview[8] * p[2] + mview[12],
                mview[1] * p[0] + mview[5] * p[1] + mview[9] * p[2] + mview[13],
                mview[2] * p[0] + mview[6] * p[1] + mview[10] * p[2] + mview[14]);

}

//finds distance between point p0 and line segment defined by points p1, p2
float PointLineSegmentDistance(const vec3 & p0, const vec3 & p1, const vec3 & p2)
{

    //project p0 onto line p1, p2
    const vec3 p1p0 = p0 - p1;
    const vec3 p1p2 = p2 - p1;
    const float p1p2len = p1p2.length();

    //get projection's parameter along line
    const float t = (p1p0 * p1p2) / p1p2len;

    if (t >= 1.0f) {
        return (p0 - p2).length();
    }
    else if (t <= 0.0f) {
        return p1p0.length();
    }
    else {

        //subtract the projection from p1p0, use that length (its the orthogonal length from p0 to line)
        return (p1p0 - (p1p2 * t)).length();

    }

}

//finds distance between point p0 and the triangle formed by points p1, p2, p3
float PointTriangleDistance(const vec3 & p0, const vec3 & p1, const vec3 & p2, const vec3 & p3)
{

    //to compute triangle interior distance, we do a ray test from p0 along the line defined by the normal of the triangle
    const vec3 ray_dir = (p2-p1) ^ (p3-p1);
    vec3 ipt;

    if (GetRayTriIntersect(p0, ray_dir, p1, p2, p3, ipt)) { //fire ray one way from p0
        return (p0 - ipt).length();
    }
    else if (GetRayTriIntersect(p0, -ray_dir, p1, p2, p3, ipt)) { //fire ray the other way from p0
        qDebug() << "::PointTriangleDistance() - HIT IN THE OTHER DIRECTION!";
        return (p0 - ipt).length();
    }
    else {

        //ray doesn't fall into triangle, an edge distance is closest
        const float dist_e0 = PointLineSegmentDistance(p0, p1, p2);
        const float dist_e1 = PointLineSegmentDistance(p0, p2, p3);
        const float dist_e2 = PointLineSegmentDistance(p0, p3, p1);

        return qMin(dist_e0, qMin(dist_e1, dist_e2));

    }

}
