#include "basrelief.h"

BasRelief::BasRelief()
{
}

void BasRelief::SetBasisVectors(const vec3 b_x, const vec3 b_y, const vec3 b_z)
{

    basis_x = b_x;
    basis_y = b_y;
    basis_z = b_z;

}

void BasRelief::SetBasisVectors(const float modelview[])
{

    basis_x = vec3(modelview[0], modelview[1], modelview[2]);
    basis_y = vec3(modelview[4], modelview[5], modelview[6]);
    basis_z = vec3(modelview[8], modelview[9], modelview[10]);

}

vec3 BasRelief::BasReliefTransformNormal(const vec3 & v, const vec3 & lmn)
{

    const float one_over_lambda = 1.0f / lmn[0];
    return vec3(v[0] - v[2] * lmn[1] * one_over_lambda,
                           v[1] - v[2] * lmn[2] * one_over_lambda,
                           v[2] * one_over_lambda);

}

float BasRelief::ComputeBasReliefL1Norm(const QList <vec3> & guess_norms, const QList <vec3> & true_norms, const vec3 & lmn)
{

    float l1_error = 0.0f;

    for (int i=0; i<guess_norms.size(); ++i) {

        //transformed guess normal
        vec3 gt = BasReliefTransformNormal(guess_norms[i], lmn);

        //normalize transformed normal!
        gt.normalize();

        //add angle error
        l1_error += AngleBetweenDeg(gt, true_norms[i]);

    }

    return l1_error;

}

void BasRelief::ComputeBasReliefTransforms(QList <StudyData> & studydatas)
{

    qDebug() << "Computing Generalized Bas-Relief Transformations on Normals...";

    QList <QList <vec3> > guess_norms;
    QList <QList <vec3> > true_norms;
    QList <QList <int> > guess_norm_index;
    QList <vec3> space_vx;
    QList <vec3> space_vy;
    QList <vec3> space_vz;

    //user's data will be organized into groups of 10, and put into the right space for the bas-relief transform

    int each_norms = 0;

    QList <vec3> each_guess_norms;
    QList <vec3> each_true_norms;
    QList <int> each_guess_norm_index;

    for (int i=0; i<studydatas.size(); ++i) {

        StudyData & d = studydatas[i];

        //use camera parameters to define worldspace
        /*
        SnapCameraToStaticView(d.model_index);

        vec3 v_y = cam.Up();
        vec3 v_z = cam.ViewDir();
        vec3 v_x = v_y ^ v_z;

        v_x.normalize();
        v_y.normalize();
        v_z.normalize();
        */

        //convert guessed norm and true norm from norms, to their endpoints in space,
        //where x, y are projected points, and z is the depth value
        vec3 g = vec3(d.guessed_normal * basis_x, d.guessed_normal * basis_y, d.guessed_normal * basis_z);
        vec3 t = vec3(d.true_normal * basis_x, d.true_normal * basis_y, d.true_normal * basis_z);

        g.normalize();
        t.normalize();

        each_guess_norms.push_back(g);
        each_true_norms.push_back(t);
        each_guess_norm_index.push_back(i);

        ++each_norms;

        /*
        if (each_norms >= study_numgaugespermodel * 2) {

            //fire it off to the lists
            guess_norms.push_back(each_guess_norms);
            true_norms.push_back(each_true_norms);
            guess_norm_index.push_back(each_guess_norm_index);

            space_vx.push_back(v_x);
            space_vy.push_back(v_y);
            space_vz.push_back(v_z);

            //clear sublists, reset counter
            each_guess_norms.clear();
            each_true_norms.clear();
            each_guess_norm_index.clear();

            each_norms = 0;

        }
        */

    }


    //now for every group of 10, we apply a convergent algorithm to find best lambda, mu, nu
    qDebug() << "Solving each system";
    for (int i=0; i<guess_norms.size(); ++i) {

        //debug show transformed guess/true normals
        //for (int j=0; j<guess_norms[i].size(); ++j) {
        //    qDebug() << i << j << guess_norms[i][j] << true_norms[i][j] << "\n";
        //}

        vec3 lmn(1.0, 0.0, 0.0);

        float epsilon = 0.001f;
        float l1_start = ComputeBasReliefL1Norm(guess_norms[i], true_norms[i], lmn);
        float l1;

        int iters = 0;

        //qDebug() << "Solving system" << i;
        do {

            //perform finite differencing in each dimension,
            l1 = ComputeBasReliefL1Norm(guess_norms[i], true_norms[i], lmn);
            float l1_dl = ComputeBasReliefL1Norm(guess_norms[i], true_norms[i], lmn + vec3(epsilon, 0, 0)); //delta lambda
            float l1_dm = ComputeBasReliefL1Norm(guess_norms[i], true_norms[i], lmn + vec3(0, epsilon, 0)); //delta mu
            float l1_dn = ComputeBasReliefL1Norm(guess_norms[i], true_norms[i], lmn + vec3(0, 0, epsilon)); //delta nu

            //then do gradient descent
            vec3 delta_lmn = vec3(l1_dl - l1, l1_dm - l1, l1_dn - l1) * epsilon;
            //qDebug() << "l1_start" << l1_start << "l1" << l1 << "delta_lmn" << delta_lmn;

            float l1_new = ComputeBasReliefL1Norm(guess_norms[i], true_norms[i], lmn - delta_lmn);

            if (isnan(l1_new)) {
                break;
            }

            ++iters;

            //if (iters % 1000 == 0) {
            //    qDebug() << iters << l1 << l1_new << l1_dl << l1_dm << l1_dn << lmn << delta_lmn;
            //}

            //if new spot improves, great, if not, break out, we've converged
            if (l1_new > l1) {
                break;
            }
            else {
                lmn = lmn - delta_lmn;
            }

        } while (true);

        qDebug() << "Iterations to solve system" << i << ":" << iters << "L1_start" << l1_start << "L1_end" << l1 << "LMN:" << lmn;

        //save that transformed normal
        for (int j=0; j<guess_norms[i].size(); ++j) {

            vec3 bas_norm = BasReliefTransformNormal(guess_norms[i][j], lmn); //get best norm in "bas-space"
            vec3 bas_norm_trans = space_vx[i] * bas_norm[0] + space_vy[i] * bas_norm[1] + space_vz[i] * bas_norm[2]; //convert back to worldspace

            bas_norm_trans.normalize();

            studydatas[guess_norm_index[i][j]].bas_corrected_normal = bas_norm_trans;

        }

    }

    //qDebug() << v_x << v_y << v_z;
    for (int i=0; i<studydatas.size(); ++i) {
        qDebug() << i << studydatas[i].bas_corrected_normal;
    }

}

void BasRelief::ComputeBasReliefTransforms(const QList <vec3> & ground_truth, const QList <vec3> & input, QList <vec3> & transformed)
{

    //qDebug() << "Computing Generalized Bas-Relief Transformations on Normals between" << ground_truth.size() << "and" << input.size() << "normal vectors";

    //these vector lists contain the normal vectors in camera space
    QList <vec3> true_norms;
    QList <vec3> guess_norms;

    for (int i=0; i<ground_truth.size(); ++i) {

        //convert guessed norm and true norm from norms, to their endpoints in space,
        //where x, y are projected points, and z is the depth value
        vec3 t = vec3(ground_truth[i] * basis_x, ground_truth[i] * basis_y, ground_truth[i] * basis_z);
        vec3 g = vec3(input[i] * basis_x, input[i] * basis_y, input[i] * basis_z);

        t.normalize();
        g.normalize();

        true_norms.push_back(t);
        guess_norms.push_back(g);

    }


    //we apply gradient descent find best lambda, mu, nu - parameters for the generalized bas relief transformation
    //we initialize lambda = 1.0 (scaling), mu and nu = 0.0 (shearing)
    vec3 lmn(1.0, 0.0, 0.0);

    const float epsilon = 0.001f;
    //const float l1_start = ComputeBasReliefL1Norm(guess_norms, true_norms, lmn);

    int iters = 0;

    //qDebug() << "Solving the system";

    do {

        //perform finite differencing in each dimension (x, y, z <-> l, m, n)
        const float l1 = ComputeBasReliefL1Norm(guess_norms, true_norms, lmn);
        const float l1_dl = ComputeBasReliefL1Norm(guess_norms, true_norms, lmn + vec3(epsilon, 0, 0)); //delta lambda
        const float l1_dm = ComputeBasReliefL1Norm(guess_norms, true_norms, lmn + vec3(0, epsilon, 0)); //delta mu
        const float l1_dn = ComputeBasReliefL1Norm(guess_norms, true_norms, lmn + vec3(0, 0, epsilon)); //delta nu

        //then do gradient descent
        const vec3 delta_lmn = vec3(l1_dl - l1, l1_dm - l1, l1_dn - l1) * epsilon;

        const float l1_new = ComputeBasReliefL1Norm(guess_norms, true_norms, lmn - delta_lmn);

        if (isnan(l1_new)) {
            break;
        }

        ++iters;

        //if (iters % 1000 == 0) {
        //    qDebug() << iters << l1 << l1_new << l1_dl << l1_dm << l1_dn << lmn << delta_lmn;
        //}

        //if new spot improves, great, if not, break out, we've converged
        if (l1_new > l1) {
            break;
        }
        else {
            lmn = lmn - delta_lmn;
        }

    } while (true);

    //const float l1_end = ComputeBasReliefL1Norm(guess_norms, true_norms, lmn);

    //qDebug() << "Iterations to solve system:" << iters << "L1_start" << l1_start << "L1_end" << l1_end << "LMN:" << lmn;

    //save that transformed normal
    for (int i=0; i<guess_norms.size(); ++i) {

        vec3 bas_norm = BasReliefTransformNormal(guess_norms[i], lmn); //get best norm in "bas-space"
        vec3 bas_norm_trans = basis_x * bas_norm[0] + basis_y * bas_norm[1] + basis_z * bas_norm[2]; //convert back to worldspace

        bas_norm_trans.normalize();

        transformed.push_back(bas_norm_trans);

    }

    //qDebug() << v_x << v_y << v_z;
    //for (int i=0; i<studydatas.size(); ++i) {
    //    qDebug() << i << studydatas[i].bas_corrected_normal;
    //}

}
