#include "colestudydata.h"

ColeStudyData::ColeStudyData()
{

    base_dir = QString("./cole_study/");

    model_names.resize(12);
    model_names[0] = "bumps";
    model_names[1] = "cervical";
    model_names[2] = "cubehole";
    model_names[3] = "femur";
    model_names[4] = "flange";
    model_names[5] = "lumpcloth";
    model_names[6] = "pulley";
    model_names[7] = "rockerarm";
    model_names[8] = "screwdriver";
    model_names[9] = "tooth";
    model_names[10] = "twoboxcloth";
    model_names[11] = "vertebra";

    gauge_scales.resize(12);
    gauge_scales[0] = 0.5f;
    gauge_scales[1] = 0.0075f;
    gauge_scales[2] = 0.5f;
    gauge_scales[3] = 0.025f;
    gauge_scales[4] = 1.2f;
    gauge_scales[5] = 0.025f;
    gauge_scales[6] = 1.0f;
    gauge_scales[7] = 0.01f;
    gauge_scales[8] = 0.001f;
    gauge_scales[9] = 0.25f;
    gauge_scales[10] = 0.04f;
    gauge_scales[11] = 0.08f;

    using_camera_a = QVector <bool> (12, true);
    using_camera_a[4] = false;
    using_camera_a[7] = false;
    using_camera_a[11] = false;

}

void ColeStudyData::Load()
{

    LoadCameraSettings();
    LoadGroundTruths();
    LoadAskOrders();
    LoadSlantTilt();
    LoadTimes();    
    LoadModels();    
    ComputeGaugeNearestVertex();

    //ComputeBasRelief();
    //SaveBasRelief();
    LoadBasRelief();

    SetupGroundTruthsGauges();
    SetupUserAverageGauges();

    LoadCurvaturesVertex();
    ComputeCurvaturesGauge();

    LoadViewDepCurvaturesVertex();
    ComputeViewDepCurvaturesGauge();

    //SaveCSVCurvatureAverageError();
    SaveCSVViewAngleAverageError();
    SaveCSVViewDepCurvatureAverageError();
    //SaveCSVMultiDimensionPlot();

}

void ColeStudyData::Draw(const int model_index)
{

    const double aspect = double(camera_settings[model_index].screen_x) / double(camera_settings[model_index].screen_y);

    glViewport(0, 0, camera_settings[model_index].screen_x, camera_settings[model_index].screen_y);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();    
    if (model_index == 8) {
        gluPerspective(30.0, aspect, 0.001, 1.0);
    }
    else {
        gluPerspective(30.0, aspect, 1.0, 400.0);
    }        

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    //glLoadMatrixf(camera_settings[model_index].modelview);

    //models[model_index].DrawGL();
    //models[model_index].DrawCurvatureGL(curvatures_vertex[model_index]);
    models[model_index].DrawCurvatureGL(viewdep_curvatures_vertex[model_index]);

    //glMatrixMode(GL_MODELVIEW);
    //glLoadIdentity();

    glClear(GL_DEPTH_BUFFER_BIT);

    //for (int i=0; i<ground_truths_gauges[model_index].size(); ++i) {
    //    ground_truths_gauges[model_index][i].DrawLineGL();
    //}

    for (int i=0; i<user_avg_gauges_BR[model_index].size(); ++i) {
        if (ground_truths[model_index][i].group == 2) {
            user_avg_gauges_BR[model_index][i].DrawGL(1.0f);
        }
    }

}

int ColeStudyData::GetViewportWidth(const int model_index)
{

    return camera_settings[model_index].screen_x;

}

int ColeStudyData::GetViewportHeight(const int model_index)
{

    return camera_settings[model_index].screen_y;

}

void ColeStudyData::SlantTiltToNormal(const float slant, const float tilt, vec3 & normal)
{

    const float theta = (90.0f - tilt) * DEG_TO_RAD;
    const float phi = (slant) * DEG_TO_RAD;

    normal[0] = cosf(theta) * sinf(phi);
    normal[1] = sinf(theta) * sinf(phi);
    normal[2] = cosf(phi);
    normal.normalize();

}

void ColeStudyData::NormalToSlantTilt(const vec3 & normal, float & slant, float & tilt)
{

    const float phi = acosf(normal[2]) * RAD_TO_DEG;
    const float theta = atan2f(normal[1], normal[0]) * RAD_TO_DEG;

    slant = phi;
    tilt = (90.0f - theta) + 360 * (theta > 90.0f);

}

void ColeStudyData::LoadGroundTruths()
{

    ground_truths.clear();
    ground_truths.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/stim/") + model_names[i] + QString("/gauges.txt");

        qDebug() << "Loading" << filename << "...";
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadGroundTruths() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        while (!ifs.atEnd()) {

            QStringList eachline = ifs.readLine().split(" ");

            if (eachline.size() < 12) {
                qDebug() << "ColeStudyData::LoadGroundTruths() File " << filename << " has invalid format, line has" << eachline.size() << "entries.";
                file.close();
                return;
            }

            GroundTruth each;
            each.index = eachline[0].toInt();
            each.group = eachline[1].toInt();
            each.screen_x = eachline[2].toInt();
            each.screen_y = eachline[3].toInt();
            each.slant = eachline[4].toFloat();
            each.tilt = eachline[5].toFloat();
            each.normal[0] = eachline[6].toFloat();
            each.normal[1] = eachline[7].toFloat();
            each.normal[2] = eachline[8].toFloat();
            each.position[0] = eachline[9].toFloat();
            each.position[1] = eachline[10].toFloat();
            each.position[2] = eachline[11].toFloat();

            ground_truths[i].push_back(each);

        }

        file.close();

    }

}

void ColeStudyData::LoadAskOrders()
{

    ask_orders.clear();
    ask_orders.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("/l/ask_orders.txt");

        qDebug() << "Loading" << filename << "...";
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadAskOrders() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        QStringList eachline = ifs.readLine().split(" ");

        if (eachline.size() < 3) {
            qDebug() << "ColeStudyData::LoadAskOrders() File " << filename << " has invalid format, line has" << eachline.size() << "entries.";
            file.close();
            return;
        }

        int n_rows = eachline[0].toInt();
        int n_cols = eachline[1].toInt();

        ask_orders[i].resize(n_rows);
        for (int j=0; j<n_rows; ++j) {
            ask_orders[i][j] = QVector <int> (n_cols, 0);
        }

        while (!ifs.atEnd()) {

            eachline = ifs.readLine().split(" ");

            if (eachline.size() < 3) {
                qDebug() << "ColeStudyData::LoadAskOrders() File " << filename << " has invalid format, line has" << eachline.size() << "entries.";
                file.close();
                return;
            }

            int row = eachline[0].toInt() - 1;
            int col = eachline[1].toInt() - 1;

            ask_orders[i][row][col] = eachline[2].toInt();

        }

        file.close();

    }

}

void ColeStudyData::LoadSlantTilt()
{

    slant_tilts.clear();
    slant_tilts.resize(model_names.size());
    slant_tilts_set.clear();
    slant_tilts_set.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("/l/slant_tilt.txt");

        qDebug() << "Loading" << filename << "...";
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadSlantTilt() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        QStringList eachline = ifs.readLine().split(" ");

        if (eachline.size() < 3) {
            qDebug() << "ColeStudyData::LoadSlantTilt() File " << filename << " has invalid format, line has" << eachline.size() << "entries.";
            file.close();
            return;
        }

        int n_rows = eachline[0].toInt();
        int n_cols = eachline[1].toInt();

        slant_tilts[i].resize(n_rows);
        slant_tilts_set[i].resize(n_rows);
        for (int j=0; j<n_rows; ++j) {
            slant_tilts[i][j] = QVector <vec2> (n_cols, vec2(0, 0));
            slant_tilts_set[i][j] = QVector <bool> (n_cols, false);
        }

        while (!ifs.atEnd()) {

            eachline = ifs.readLine().split(" ");

            if (eachline.size() < 4) {
                qDebug() << "ColeStudyData::LoadSlantTilt() File " << filename << " has invalid format, line has" << eachline.size() << "entries.";
                file.close();
                return;
            }

            int row = eachline[0].toInt() - 1;
            int col = eachline[1].toInt() - 1;

            slant_tilts[i][row][col] = vec2(eachline[2].toFloat(), eachline[3].toFloat());
            slant_tilts_set[i][row][col] = true;

        }

        file.close();

    }

}

void ColeStudyData::LoadTimes()
{

    times.clear();
    times.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("/l/times.txt");

        qDebug() << "Loading" << filename << "...";
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadTimes() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        QStringList eachline = ifs.readLine().split(" ");

        if (eachline.size() < 3) {
            qDebug() << "ColeStudyData::LoadTimes() File " << filename << " has invalid format, line has" << eachline.size() << "entries.";
            file.close();
            return;
        }

        int n_rows = eachline[0].toInt();
        int n_cols = eachline[1].toInt();

        times[i].resize(n_rows);
        for (int j=0; j<n_rows; ++j) {
            times[i][j] = QVector <int> (n_cols, 0);
        }

        while (!ifs.atEnd()) {

            eachline = ifs.readLine().split(" ");

            if (eachline.size() < 3) {
                qDebug() << "ColeStudyData::LoadTimes() File " << filename << " has invalid format, line has" << eachline.size() << "entries.";
                file.close();
                return;
            }

            int row = eachline[0].toInt() - 1;
            int col = eachline[1].toInt() - 1;

            times[i][row][col] = eachline[2].toInt();

        }

        file.close();

    }

}

void ColeStudyData::LoadModels()
{

    models.clear();
    models.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("line_study/models/") + model_names[i] + QString("_viewfix.obj");

        models[i].Read_OBJ(filename);

        qDebug() << "Read OBJ" << filename << "with" << models[i].GetVerts().size() << "verts and" << models[i].GetFaces().size() << "faces.";

        /*
        QString filename = base_dir + QString("line_study/models/") + model_names[i] + QString(".obj");

        models[i].Read_OBJ(filename);        

        qDebug() << "Read OBJ" << filename << "with" << models[i].GetVerts().size() << "verts and" << models[i].GetFaces().size() << "faces.";

        QString out_filename = base_dir + QString("line_study/models/") + model_names[i] + QString("_viewfix.obj");

        models[i].ApplyModelviewMatrix(camera_settings[i].modelview);
        //models[i].FlipNormals();
        models[i].Write_OBJ(out_filename);

        qDebug() << "Wrote OBJ" << out_filename << "with" << models[i].GetVerts().size() << "verts and" << models[i].GetFaces().size() << "faces.";
        */

    }

}

void ColeStudyData::LoadCameraSettings()
{

    camera_settings.clear();
    camera_settings.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename;

        if (using_camera_a[i]) {
            filename = base_dir + QString("line_study/cameras/") + model_names[i] + QString("_a.xf");
        }
        else {
            filename = base_dir + QString("line_study/cameras/") + model_names[i] + QString("_b.xf");
        }

        qDebug() << "Loading" << filename << "...";
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadCameraSettings() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        QStringList eachline;
        CameraSetting c;

        for (int j=0; j<4; ++j) {

            eachline = ifs.readLine().split(" ");

            c.modelview[0 + j] = eachline[0].toFloat();
            c.modelview[4 + j] = eachline[1].toFloat();
            c.modelview[8 + j] = eachline[2].toFloat();
            c.modelview[12 + j] = eachline[3].toFloat();

        }

        eachline = ifs.readLine().split(" ");
        c.fov = eachline[0].toFloat();

        eachline = ifs.readLine().split(" ");
        c.screen_x = eachline[0].toInt();
        c.screen_y = eachline[1].toInt();

        camera_settings[i] = c;

        file.close();

    }

}

void ColeStudyData::LoadCurvaturesVertex()
{

    curvatures_vertex.clear();
    curvatures_vertex.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("/l/out_curv.txt");

        qDebug() << "Loading" << filename << "...";
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadCurvatures() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        while (!ifs.atEnd()) {

            QStringList eachline = ifs.readLine().split(" ");

            curvatures_vertex[i].push_back(vec2(eachline[1].toFloat(), eachline[2].toFloat()));

        }

        file.close();

        if (curvatures_vertex[i].size() != models[i].GetVerts().size()) {
            qDebug() << "ColeStudyData::LoadCurvatures() PROBLEM!  Entries in curvature file"
                     << curvatures_vertex[i].size()
                     << "does not match vertex count for model"
                     << model_names[i]
                     << models[i].GetVerts().size();
        }

    }

}

void ColeStudyData::ComputeCurvaturesGauge()
{

    //estimate curvature for the gauges by using curvature at closest vertex
    curvatures_gauges.clear();
    curvatures_gauges.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        for (int j=0; j<ground_truths[i].size(); ++j) {

            const int vert_index = gauge_nearest_vertex[i][j];

            curvatures_gauges[i].push_back(curvatures_vertex[i][vert_index]);

        }

    }

}

void ColeStudyData::LoadViewDepCurvaturesVertex()
{

    viewdep_curvatures_vertex.clear();
    viewdep_curvatures_vertex.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("/l/out_viewdep_curv.txt");

        qDebug() << "Loading" << filename << "...";
        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadViewDepCurvaturesVertex() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        while (!ifs.atEnd()) {

            QStringList eachline = ifs.readLine().split(" ");

            viewdep_curvatures_vertex[i].push_back(eachline[1].toFloat());

        }

        file.close();

        if (viewdep_curvatures_vertex[i].size() != models[i].GetVerts().size()) {
            qDebug() << "ColeStudyData::LoadViewDepCurvaturesVertex() PROBLEM!  Entries in viewdep curvature file"
                     << viewdep_curvatures_vertex[i].size()
                     << "does not match vertex count for model"
                     << model_names[i]
                     << models[i].GetVerts().size();
        }

    }

}

void ColeStudyData::ComputeViewDepCurvaturesGauge()
{

    //estimate curvature for the gauges by using curvature at closest vertex
    viewdep_curvatures_gauges.clear();
    viewdep_curvatures_gauges.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        for (int j=0; j<ground_truths[i].size(); ++j) {

            const int vert_index = gauge_nearest_vertex[i][j];

            viewdep_curvatures_gauges[i].push_back(viewdep_curvatures_vertex[i][vert_index]);

        }

    }

}

void ColeStudyData::SetupGroundTruthsGauges()
{

    ground_truths_gauges.clear();
    ground_truths_gauges.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        for (int j=0; j<ground_truths[i].size(); ++j) {

            GaugeManipulator g;

            g.SetCentre(ground_truths[i][j].position);
            g.SetNormal(ground_truths[i][j].normal);
            g.SetScale(gauge_scales[i]);

            ground_truths_gauges[i].push_back(g);

        }

    }

}

void ColeStudyData::SetupUserAverageGauges()
{

    user_avg_gauges.clear();
    user_avg_gauges.resize(model_names.size());

    user_avg_gauges_BR.clear();
    user_avg_gauges_BR.resize(model_names.size());

    avg_error_degrees.clear();
    avg_error_degrees.resize(model_names.size());

    avg_error_degrees_BR.clear();
    avg_error_degrees_BR.resize(model_names.size());

    gauge_avg_error.clear();
    gauge_avg_error.resize(model_names.size());

    gauge_avg_error_BR.clear();
    gauge_avg_error_BR.resize(model_names.size());

    //for each model
    for (int i=0; i<model_names.size(); ++i) {

        int grand_total_err_gauges = 0;
        float grand_total_err = 0.0f;
        float grand_total_err_BR = 0.0f;

        //for each "ground truth" gauge in the set
        for (int j=0; j<ground_truths[i].size(); ++j) {

            //vec3 truth_norm = ground_truths[i][j].normal;
            vec3 truth_norm;
            SlantTiltToNormal(ground_truths[i][j].slant, ground_truths[i][j].tilt, truth_norm);

            //iterate over all users who did this gauge (think it's a row in a matrix?)
            int n_users = 0;
            float total_err = 0.0f;
            float total_err_BR = 0.0f;

            vec3 avg_norm = vec3(0, 0, 0);
            vec3 avg_norm_BR = vec3(0, 0, 0);

            for (int k=0; k<slant_tilts[i][j].size(); ++k) {

                if (slant_tilts_set[i][j][k]) {

                    //add to average error
                    vec3 each_norm;
                    SlantTiltToNormal(slant_tilts[i][j][k][0], slant_tilts[i][j][k][1], each_norm);
                    total_err += acosf(each_norm * truth_norm) * RAD_TO_DEG;
                    avg_norm = avg_norm + each_norm;

                    //add to average error (BR)
                    vec3 each_norm_BR = input_normals_BR[i][j][k];
                    total_err_BR += acosf(each_norm_BR * truth_norm) * RAD_TO_DEG;
                    avg_norm_BR = avg_norm_BR + each_norm_BR;

                    ++n_users;

                }

            }

            //save average gauge error
            const float gauge_error = total_err / float(n_users);
            const float gauge_error_BR = total_err_BR / float(n_users);

            gauge_avg_error[i].push_back(gauge_error);
            gauge_avg_error_BR[i].push_back(gauge_error_BR);

            //now compute gauge stuff
            avg_norm = avg_norm / float(n_users);
            avg_norm_BR = avg_norm_BR / float(n_users);

            GaugeManipulator g;

            g.SetCentre(ground_truths[i][j].position);
            g.SetNormal(avg_norm);
            g.SetScale(gauge_scales[i]);
            g.SetColourByError(gauge_error);

            user_avg_gauges[i].push_back(g);

            GaugeManipulator g_br;

            g_br.SetCentre(ground_truths[i][j].position);
            g_br.SetNormal(avg_norm_BR);
            g_br.SetScale(gauge_scales[i]);
            g_br.SetColourByError(gauge_error_BR);

            user_avg_gauges_BR[i].push_back(g_br);

            if (ground_truths[i][j].group == 0) {

                grand_total_err += total_err / float(n_users);
                grand_total_err_BR += total_err_BR / float(n_users);
                ++grand_total_err_gauges;

            }

        }

        avg_error_degrees[i] = grand_total_err / float(grand_total_err_gauges);
        avg_error_degrees_BR[i] = grand_total_err_BR / float(grand_total_err_gauges);
    }

    for (int i=0; i<model_names.size(); ++i) {
        qDebug() << "Model" << i << "(" << model_names[i] << ") - average error " << avg_error_degrees[i] << "degrees";
    }
    for (int i=0; i<model_names.size(); ++i) {
        qDebug() << "Model" << i << "(" << model_names[i] << ") - average error BR-corrected " << avg_error_degrees_BR[i] << "degrees";
    }

    /*
    QString slant_str1;
    QString slant_str2;
    QString slant_str3;

    vec3 truth_norm;
    SlantTiltToNormal(ground_truths[4][0].slant, ground_truths[4][0].tilt, truth_norm);

    for (int i=0; i<slant_tilts[4][0].size(); ++i) {
        if (slant_tilts_set[4][0][i]) {
            slant_str1 += QString::number(slant_tilts[4][0][i][0]) + QString(" ");
            slant_str2 += QString::number(slant_tilts[4][0][i][1]) + QString(" ");

            vec3 each_norm;
            SlantTiltToNormal(slant_tilts[4][0][i][0], slant_tilts[4][0][i][1], each_norm);
            slant_str3 += QString::number(acosf(each_norm * truth_norm) / ALG_M_PI_2) + QString(" ");
        }
    }

    qDebug() << slant_str1;
    qDebug() << slant_str2;
    qDebug() << slant_str3;
    qDebug() << ground_truths[4][0].slant << ground_truths[4][0].tilt;
    */

}

void ColeStudyData::ComputeBasRelief()
{

    qDebug() << "Computing Bas Relief transforms for all input...";

    //Bas relief needs to be computed per-participant (assumption: ambiguity varies by participant)

    input_normals_BR.clear();
    input_normals_BR.resize(model_names.size());

    //index i - MODEL INDEX
    //index j - GAUGE INDEX
    //index k - PARTICIPANT INDEX (note: participant K may not have set gauge J for model I)

    //first resize
    for (int i=0; i<model_names.size(); ++i) {

        input_normals_BR[i].resize(ground_truths[i].size());

        for (int j=0; j<ground_truths[i].size(); ++j) {
            input_normals_BR[i][j].resize(slant_tilts[i][j].size());
        }

    }

    //then for each model
    for (int i=0; i<model_names.size(); ++i) {

        int n_rows = slant_tilts[i].size();
        int n_cols = slant_tilts[i][0].size();

        //for each participant
        for (int k=0; k<n_cols; ++k) {

            //setup bas relief structure
            BasRelief basRelief;
            basRelief.SetBasisVectors(camera_settings[i].modelview);

            QList <vec3> gts;
            QList <vec3> inputs;
            QList <vec3> inputs_br;

            //for each gauge
            for (int j=0; j<n_rows; ++j) {

                //did participant k set gauge j for model i?
                if (slant_tilts_set[i][j][k]) {

                    vec3 truth_norm;
                    SlantTiltToNormal(ground_truths[i][j].slant, ground_truths[i][j].tilt, truth_norm);
                    gts.push_back(truth_norm);

                    vec3 input_norm;
                    SlantTiltToNormal(slant_tilts[i][j][k][0], slant_tilts[i][j][k][1], input_norm);
                    inputs.push_back(input_norm);

                }

            }

            //participant didn't work on this
            if (gts.empty() || inputs.empty()) {
                continue;
            }

            //compute the BRT
            basRelief.ComputeBasReliefTransforms(gts, inputs, inputs_br);

            //now set corresponding transformed norms for each i,j,k
            int inputs_br_index = 0;
            for (int j=0; j<n_rows; ++j) {

                //did participant k set gauge j for model i?
                if (slant_tilts_set[i][j][k]) {

                    input_normals_BR[i][j][k] = inputs_br[inputs_br_index];
                    ++inputs_br_index;

                }

            }

        }

    }        

}

void ColeStudyData::SaveBasRelief()
{

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("/l/norms_bas_corrected.txt");

        qDebug() << "Saving" << filename << "...";

        QFile file(filename);

        if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::SaveBasRelief() File " << filename << " can't be opened for writing.";
            return;
        }

        QTextStream ofs(&file);

        ofs << input_normals_BR[i].size() << " " << input_normals_BR[i][0].size() << " 0\n";
        for (int k=0; k<input_normals_BR[i][0].size(); ++k) {
            for (int j=0; j<input_normals_BR[i].size(); ++j) {

                if (slant_tilts_set[i][j][k]) {
                    ofs << j+1 << " " << k+1 << " " << input_normals_BR[i][j][k][0] << " " << input_normals_BR[i][j][k][1] << " " << input_normals_BR[i][j][k][2] << "\n";
                }

            }

        }

        file.close();

    }

}

void ColeStudyData::LoadBasRelief()
{

    input_normals_BR.clear();
    input_normals_BR.resize(model_names.size());

    //index i - MODEL INDEX
    //index j - GAUGE INDEX
    //index k - PARTICIPANT INDEX (note: participant K may not have set gauge J for model I)

    //first resize
    for (int i=0; i<model_names.size(); ++i) {

        input_normals_BR[i].resize(ground_truths[i].size());

        for (int j=0; j<ground_truths[i].size(); ++j) {
            input_normals_BR[i][j].resize(slant_tilts[i][j].size());
        }

    }

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("/l/norms_bas_corrected.txt");

        qDebug() << "Loading" << filename << "...";

        QFile file(filename);

        if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::LoadBasRelief() File " << filename << " can't be opened for reading.";
            return;
        }

        QTextStream ifs(&file);

        //skip past the header bit (dimensions of the array)
        ifs.readLine();

        while (!ifs.atEnd()) {

            QStringList eachline = ifs.readLine().split(" ");

            if (eachline.size() < 5) {
                qDebug() << "ColeStudyData::LoadBasRelief() File " << filename << " has invalid format.";
                return;
            }

            const int j = eachline[0].toInt() - 1;
            const int k = eachline[1].toInt() - 1;

            input_normals_BR[i][j][k] = vec3(eachline[2].toFloat(), eachline[3].toFloat(), eachline[4].toFloat());
            input_normals_BR[i][j][k].normalize();

        }

        file.close();

    }

}

void ColeStudyData::ComputeGaugeNearestVertex()
{

    gauge_nearest_vertex.clear();
    gauge_nearest_vertex.resize(model_names.size());

    for (int i=0; i<model_names.size(); ++i) {

        //note: assumed modelview matrix already applied to vertices so that model vertices/gauge positions line up

        for (int j=0; j<ground_truths[i].size(); ++j) {

            //note!  each model has already been transformed by the modelview matrix, so we can use the actual gauge position
            int vert_ind = models[i].GetClosestVertexIndex(ground_truths[i][j].position);

            gauge_nearest_vertex[i].push_back(vert_ind);

        }


    }

}

void ColeStudyData::SaveCSVCurvatureAverageError()
{

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("_curvature_avgerror.csv");

        qDebug() << "Saving" << filename << "...";

        QFile file(filename);

        if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::SaveCSVCurvatureAverageError() File " << filename << " can't be opened for writing.";
            return;
        }

        QTextStream ofs(&file);

        ofs << "GaugeIndex,NearestVertex,AbsoluteMeanCurvature,AbsoluteGaussianCurvature,AbsoluteKappa1,GaugeAverageError\n";

        for (int j=0; j<ground_truths[i].size(); ++j) {

            const float mean_curv = abs(curvatures_gauges[i][j][0] + curvatures_gauges[i][j][1]) * 0.5f;
            const float gauss_curv = abs(curvatures_gauges[i][j][0] * curvatures_gauges[i][j][1]);
            const float kappa_1 = abs(curvatures_gauges[i][j][0]);

            ofs << (j+1) << "," << gauge_nearest_vertex[i][j] << "," << mean_curv << "," << gauss_curv << "," << kappa_1 << "," << gauge_avg_error[i][j] << "\n";

        }

        file.close();

    }

}

void ColeStudyData::SaveCSVViewAngleAverageError()
{

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("_viewangle_avgerror.csv");

        qDebug() << "Saving" << filename << "...";

        QFile file(filename);

        if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::SaveCSVViewAngleAverageError() File " << filename << " can't be opened for writing.";
            return;
        }

        QTextStream ofs(&file);

        ofs << "GaugeIndex,ViewAngle,GaugeAverageError\n";

        for (int j=0; j<ground_truths[i].size(); ++j) {

            const float angle1 = AngleBetweenDeg(vec3(0, 0, 1), ground_truths[i][j].normal);
            const float angle2 = AngleBetweenDeg(vec3(0, 0, -1), ground_truths[i][j].normal);
            const float view_angle = qMin(angle1, angle2);

            if (ground_truths[i][j].group == 2) {
                ofs << (j+1) << "," << view_angle << "," << gauge_avg_error[i][j] << "\n";
            }

        }

        file.close();

    }

}

void ColeStudyData::SaveCSVViewDepCurvatureAverageError()
{

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("_viewdepcurv_avgerror.csv");

        qDebug() << "Saving" << filename << "...";

        QFile file(filename);

        if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::SaveCSVViewAngleAverageError() File " << filename << " can't be opened for writing.";
            return;
        }

        QTextStream ofs(&file);

        ofs << "GaugeIndex,ViewDepCurvature(q1),GaugeAverageError\n";

        for (int j=0; j<ground_truths[i].size(); ++j) {

            if (ground_truths[i][j].group == 2) {
                ofs << (j+1) << "," << viewdep_curvatures_vertex[i][j] << "," << gauge_avg_error[i][j] << "\n";
            }

        }

        file.close();

    }

}

void ColeStudyData::SaveCSVMultiDimensionPlot()
{

    for (int i=0; i<model_names.size(); ++i) {

        QString filename = base_dir + QString("gauge_settings/data/") + model_names[i] + QString("_multidimension.csv");

        qDebug() << "Saving" << filename << "...";

        QFile file(filename);

        if (!file.open(QIODevice::WriteOnly | QIODevice::Text)) {
            qDebug() << "ColeStudyData::SaveCSVViewAngleAverageError() File " << filename << " can't be opened for writing.";
            return;
        }

        QTextStream ofs(&file);

        for (int k=0; k<2; ++k) {

            if (k == 0) {
                ofs << "GaugeIndex,ViewAngle,AbsoluteKappa1,GaugeAverageErrorLOW\n";
            }
            else {
                ofs << "GaugeIndex,ViewAngle,AbsoluteKappa1,GaugeAverageErrorHIGH\n";
            }

            for (int j=0; j<ground_truths[i].size(); ++j) {

                if (k == 0 && gauge_avg_error[i][j] > 30.0f) {
                    continue;
                }
                else if (k == 1 && gauge_avg_error[i][j] <= 30.0f) {
                    continue;
                }

                const float kappa_1 = abs(curvatures_gauges[i][j][0]);

                const float angle1 = AngleBetweenDeg(vec3(0, 0, 1), ground_truths[i][j].normal);
                const float angle2 = AngleBetweenDeg(vec3(0, 0, -1), ground_truths[i][j].normal);
                const float view_angle = qMin(angle1, angle2);

                ofs << (j+1) << "," << view_angle << "," << kappa_1 << "," << gauge_avg_error[i][j] << "\n";

            }

        }

        file.close();

    }

}
