#ifndef GLWIDGET_H
#define GLWIDGET_H

#include <QtCore>
#include <QtGui>
#include <QtOpenGL>

#include <cfloat>

#include "orbitcamera.h"
#include "OFFFile.h"
#include "glutils.h"
#include "gaugemanipulator.h"
#include "medialaxis.h"
#include "basrelief.h"
#include "studydata.h"
#include "turkworkerdata.h"
#include "trialrecord.h"
#include "histogram.h"
#include "scatterplot.h"

//#include "colestudydata.h"

class GLWidget : public QGLWidget
{

    Q_OBJECT

public:

    GLWidget();

    void SetupStudy(const QString & turker_id);

    QSize minimumSizeHint() const;
    QSize sizeHint() const;

    void keyPressEvent(QKeyEvent * event);
    void keyReleaseEvent(QKeyEvent * event);

    int GetCurTrial();
    int GetNumTrials();

    QString GetStudyData();
    void LoadTurkStudyDataFromCSV(const QString & filename, TurkWorkerData & datas);

    void LoadCurvatures(const QString & filename, QList <float> & curvs);
    float GetCurvatureForFace(const int model_index, const int face_index);   

    bool StudyFinished();
    bool StudySuccessful();            

signals:

    void StartedNewTrial();
    void EnableNextTrialButton();
    void DisableNextTrialButton();

public slots:

    void CopyStudyDataToClipboard();
    void SetupNewTrial();

protected:

    void initializeGL();
    void paintGL();
    void resizeGL(int width, int height);
    void mousePressEvent(QMouseEvent *event);
    void mouseMoveEvent(QMouseEvent *event);
    void mouseReleaseEvent(QMouseEvent *event);
    void wheelEvent(QWheelEvent *);    

private:

    //interface related - camera
    void UpdateCamera();
    void SnapCameraToStaticView(int model_index);
    void ZoomIn();
    void ZoomOut();    

    void LoadSplashImage();
    void DrawSplashImage();

    //utility - study setup
    void SetupAllTrials();
    void SetupTurkStudyDataGauges();

    //utility - statistics-related stuff
    float GetMedian(QList <float> & samples);
    float GetMean(QList <float> & samples);
    float GetStandardDeviation(QList <float> & samples, float mean);
    float GetNormalCurveArea(const float z);
    float GetTwoSampleTwoTailZTestPValue(QList <float> & samples_i, QList <float> & samples_j);

    //utility - randomization
    QList <int> GetRandomIndexes(int n);
    int GetRandomVisibleFace(const int mesh_index);    
    //utility - export geometry
    void ExportErrorGeometry();

    //analysis
    void DoAllHistogramAnalysis(QList <StudyData> & studydatas);
    bool EvaluatePrecisionOfStudyData();
    //analysis - generalized bas relief transform stuff
    void ComputeBasReliefTransforms(QList <StudyData> & studydatas);
    float ComputeBasReliefL1Norm(const QList <vec3> & guess_norms, const QList <vec3> & true_norms, const vec3 & lmn);
    vec3 BasReliefTransformNormal(const vec3 & v, const vec3 & lmn);
    //analysis - medial axis distance
    void ComputeMedialAxisData();
    //analysis - curvature
    void ComputeCurvatureData();
    //analysis - surface to abstraction angle difference
    void ComputeAbstractionAngleDifferences();
    //analysis - surface to abstraction distance
    void ComputeAbstractionDistances();
    void VoxelListFromSurface(const OFFFile & model, const float voxel_size, QList <vec3> & voxel_pts);
    int GetClosestPoint(const QList <vec3> & pts, const vec3 p);
    //analysis - local thickness
    void ComputeLocalThickness();
    //analysis - p-values between different abstractions for each model
    void ComputePValueTables();
    void OutputHTMLTable(QTextStream & os, const QVector <QList <float> > all_values);
    //analysis - centroid distances
    void ComputeCentroidDistances();
    //analysis - view-truenorm angle
    void ComputeViewTrueNormAngle();

    //relating to creating the ground truth gauge sequences
    float GetHaltonSequence(const int index, const int base);
    bool GetGaugeForPoint(const QPoint & point, const int mesh_index, StudyData & d);
    void ComputeHaltonSequences();
    void WriteHaltonSequences();
    bool ClosestFaceAlongRay(const int mesh_index, const vec2 & screen_pt, int & face_index, vec3 & int_pt);
    void LoadHaltonSequences();
    void AddCustomGauge(const int mesh_index);

    //members

    OrbitCamera cam;                
    GaugeManipulator gaugemanip;

    //key experiment conditions: task and abstraction
    //int participant_hit_number; //NOTE!  this defines study_task, abstraction, etc
    QString turk_worker_id;
    int hashed_id;

    TrialRecord trial_records;

    StudyTask study_task;
    int abstraction;

    //ground truth Halton sequenced gauges
    QVector <QList <StudyData> > ground_truths;
    QVector <QList <GaugeManipulator> > ground_truths_gauges;

    //instructions image
    QImage splash_image_gl;

    //keyboard+mouse stuff...
    bool shift_key_held;
    bool control_key_held;
    QPoint mouse_pos;
    QPoint mouse_saved_pos;
    bool mouse_left;
    bool mouse_middle;
    bool mouse_right;

    QTimer animate_timer;

    //study stuff   

    QList <vec3> ray1; //for debugging
    QList <vec3> ray2; //for debugging

    //meshes/abstractions - and related properties
    QList <QString> meshnames;
    QVector <OFFFile> meshes;    
    QVector <QList <float> > meshcurv1s;
    QVector <QList <float> > meshcurv2s;
    QVector <OrbitCamera> meshstaticviews;

    QList <int> study_trialmodels;
    QList <int> study_trialfaces;

    QTime study_trial_time;        

    int study_numgaugespermodel;
    int study_trialindex;
    bool study_finished;
    bool study_successful;        
    float gauge_scale;   

    int num_halton_sequence_gauges;
    int num_total_sequence_gauges;

    QList <StudyData> studydatas; //data in memory from one run of the study (this app)

    //amazon mech turk study data load/display stuff    
    TurkWorkerData worker_data;
    QList <StudyData> turk_studydatas;
    QList <GaugeManipulator> turk_studydata_gauges;

    bool draw_all_gauges;
    bool experimental_mode;
    bool load_and_show_study_data;

    //turk study data

    //analysis - medial axis distance
    //MedialAxis medialaxis;

    BasRelief basRelief;

    //ColeStudyData coleStudyData;
    //int coleStudyIndex;

};

#endif // GLWIDGET_H
