#include "medialaxis.h"

MedialAxis::MedialAxis()
{
}

vec3 MedialAxis::VoxelToSpace(const VoxelPos & p)
{
    return bbox_min + ((vec3(p.i, p.j, p.k) - vec3(1, 1, 1)) * epsilon);
}

VoxelPos MedialAxis::SpaceToVoxel(const vec3 & v)
{

    vec3 v_proj = (v - bbox_min) / epsilon + vec3(1, 1, 1);
    return VoxelPos(v_proj[0], v_proj[1], v_proj[2]);

}

QList <vec3> MedialAxis::SurfaceListToVec(const QList <VoxelPos> & s_list)
{

    QList <vec3> vec_list;

    for (int i=0; i<s_list.size(); ++i) {
        vec_list.push_back(VoxelToSpace(s_list[i]) + vec3(epsilon * 0.5f, epsilon * 0.5f, epsilon * 0.5f));
    }

    return vec_list;

}

QList <VoxelPos> MedialAxis::GetSurfaceList(const QVector <vec3> & verts, const QVector <QVector <int> > & faces, const float eps)
{

    surface_list.clear();

    if (eps <= 0.0f) {
        qDebug() << "Epsilon must be positive!  Doing nothing";
        return surface_list;
    }

    //compute bounding box
    qDebug() << "Computing bounding box...";
    bbox_min = vec3(DBL_MAX, DBL_MAX, DBL_MAX);
    bbox_max = vec3(-DBL_MAX, -DBL_MAX, -DBL_MAX);

    for (int i=0; i<verts.size(); ++i) {

        for (int j=0; j<3; ++j) {
            bbox_min[j] = qMin(verts[i][j], bbox_min[j]);
            bbox_max[j] = qMax(verts[i][j], bbox_max[j]);
        }

    }

    //set up voxel grid based on epsilon
    qDebug() << "Setting up voxel grid...";
    epsilon = eps;

    voxels_x = int((bbox_max[0] - bbox_min[0]) / epsilon) + 3;
    voxels_y = int((bbox_max[1] - bbox_min[1]) / epsilon) + 3;
    voxels_z = int((bbox_max[2] - bbox_min[2]) / epsilon) + 3;

    voxels.clear();
    voxels.resize(voxels_x);

    for (int i=0; i<voxels_x; ++i) {

        voxels[i].resize(voxels_y);

        for (int j=0; j<voxels_y; ++j) {

            voxels[i][j].resize(voxels_z);

        }
    }

    //"triangle rasterization":
    //now we need to iterate through each traingle and "rasterize" it into this 3D volume
    //so our trick here is to sweep lines across each tri, grid like, moving by epsilon
    qDebug() << "Rasterizing triangles into grid...";

    for (int i=0; i<faces.size(); ++i) {

        for (int j=2; j<faces[i].size(); ++j) {

            const vec3 & v1 = verts[faces[i][0]];
            const vec3 & v2 = verts[faces[i][j-1]];
            const vec3 & v3 = verts[faces[i][j]];

            const float v3v2len = (v3 - v2).length();
            const vec3 dir23 = (v3 - v2) / v3v2len; //normalized direction from v2 to v3

            for (float y=0.0f; y<v3v2len; y += epsilon*0.5f) {

                const vec3 v4 = v2 + (dir23 * y); //line from v1 to v4 (v4 lies between v2, v3)
                float v1v4len = (v4 - v1).length();
                const vec3 dir14 = (v4 - v1) / v1v4len;

                for (float x=0.0f; x<v1v4len; x += epsilon*0.5f) {

                    //we want to interpolate x along a line from v1 to v4
                    vec3 v = v1 + (dir14 * x);
                    VoxelPos p = SpaceToVoxel(v);

                    if (!voxels[p.i][p.j][p.k].is_surface) {
                        voxels[p.i][p.j][p.k].is_surface = true;
                        surface_list.push_back(p);
                    }

                }
            }

        }

    }
    qDebug() << "Surface intersects" << surface_list.size() << "of" << voxels_x * voxels_y * voxels_z << "voxels";

    return surface_list;

}

void MedialAxis::CreateFromSurface(const QVector <vec3> & verts, const QVector <QVector <int> > & faces, const float eps)
{   

    //do 3d triangle rasterization to get voxels
    GetSurfaceList(verts, faces, eps);

    if (surface_list.empty()) {
        qDebug() << "No points in surface list, returning...";
        return;
    }

    //do a "flood fill" to determine which are exterior voxels (we don't test medial axis on these)
    qDebug() << "Flooding voxels to determine which are outside...";
    int num_outside = 0;

    QList <VoxelPos> flood_list;
    flood_list.push_back(VoxelPos(0, 0, 0));
    voxels[0][0][0].black = false;

    while (!flood_list.empty()) {

        VoxelPos p = flood_list.first();
        flood_list.pop_front();

        ++num_outside;

        //check all neighbours
        VoxelPos n;
        for (int i=-1; i<=1; ++i) {

            n.i = p.i + i;

            if (n.i < 0 || n.i >= voxels_x) {
                continue;
            }

            for (int j=-1; j<=1; ++j) {

                n.j = p.j + j;

                if (n.j < 0 || n.j >= voxels_y) {
                    continue;
                }

                for (int k=-1; k<=1; ++k) {

                    n.k = p.k + k;

                    if (n.k < 0 || n.k >= voxels_z) {
                        continue;
                    }

                    //dont check diagonally - just the adjacent 6
                    if ((i == 0 && j == 0) ||
                            (i == 0 && k == 0) ||
                            (j == 0 && k == 0)) {

                        Voxel & vox = voxels[n.i][n.j][n.k];

                        //if neighbours not yet outside and not surface, make it outside
                        if (!vox.is_outside && !vox.is_surface) {

                            vox.is_outside = true;
                            flood_list.push_back(n);

                        }

                    }

                }
            }
        }

    }

    qDebug() << "Outside:" << num_outside << "of" << voxels_x * voxels_y * voxels_z << "voxels";
    qDebug() << "Inside:" << voxels_x * voxels_y * voxels_z - num_outside - surface_list.size();

    //true test
    int true_outside = 0;
    int true_surface = 0;
    for (int i=0; i<voxels_x; ++i) {
        for (int j=0; j<voxels_y; ++j) {
            for (int k=0; k<voxels_z; ++k) {
                Voxel & vox = voxels[i][j][k];

                if (vox.is_outside) {
                    ++true_outside;
                }
                if (vox.is_surface) {
                    ++true_surface;
                }

                if (vox.is_outside && vox.is_surface) {
                    qDebug() << "PROBLEM!";
                }
            }
        }
    }

    qDebug() << "True counts.  Outside:" << true_outside << "Surface:" << true_surface << "Total:" << true_outside + true_surface;

    //call surface+interior black
    for (int i=0; i<voxels_x; ++i) {
        for (int j=0; j<voxels_y; ++j) {
            for (int k=0; k<voxels_z; ++k) {

                Voxel & vox = voxels[i][j][k];

                if (!vox.is_outside) {

                    vox.black = true;

                }

            }
        }
    }

    //now simplify/shrink it
    qDebug() << "Shrinking/collapsing inside voxels to create medial axis...";

    qDebug() << "Using Palagyi method";
    DoPalagyi();

    //qDebug() << "Using outside distance method";
    //DoDistanceMethod();

    //qDebug() << "Using sweepy method";
    //DoSweepy();

    //store remaining voxels that are medial in medial list
    medial_list.clear();
    for (int i=0; i<voxels_x; ++i) {
        for (int j=0; j<voxels_y; ++j) {
            for (int k=0; k<voxels_z; ++k) {

                Voxel & vox = voxels[i][j][k];

                if (vox.black) {

                    medial_list.push_back(VoxelPos(i, j, k));

                }

            }
        }
    }

    //
    qDebug() << "Done.";    

}

void MedialAxis::WriteSurface(const QString & filename)
{

    surface_geom.Clear();

    for (int i=0; i<surface_list.size(); ++i) {

        vec3 v = VoxelToSpace(surface_list[i]);
        surface_geom.AddCube(v, epsilon);

    }

    surface_geom.ComputeFaceNormals();
    surface_geom.Write_OBJ(filename);

}

void MedialAxis::WriteMedialAxis(const QString & filename)
{

    medial_geom.Clear();

    for (int i=0; i<medial_list.size(); ++i) {

        vec3 v = VoxelToSpace(medial_list[i]);
        medial_geom.AddCube(v, epsilon);

    }

    medial_geom.ComputeFaceNormals();
    medial_geom.Write_OBJ(filename);

}

void MedialAxis::DrawGL()
{

    //surface_geom.DrawGL();
    medial_geom.DrawGL();

}

bool MedialAxis::IsVoxelInsideMedial(const VoxelPos & p)
{


    if (p.i > 0 && !voxels[p.i-1][p.j][p.k].black) {
        return false;
    }

    if (p.j > 0 && !voxels[p.i][p.j-1][p.k].black) {
        return false;
    }

    if (p.k > 0 && !voxels[p.i][p.j][p.k-1].black) {
        return false;
    }

    if (p.i < voxels_x && !voxels[p.i+1][p.j][p.k].black) {
        return false;
    }

    if (p.j < voxels_y && !voxels[p.i][p.j+1][p.k].black) {
        return false;
    }

    if (p.k < voxels_z && !voxels[p.i][p.j][p.k+1].black) {
        return false;
    }

    return true;

    /*
    VoxelPos n;

    for (int i=-1; i<=1; ++i) {

        n.i = p.i + i;

        if (n.i < 0 || n.i >= voxels_x) {
            continue;
        }

        for (int j=-1; j<=1; ++j) {

            n.j = p.j + j;

            if (n.j < 0 || n.j >= voxels_y) {
                continue;
            }

            for (int k=-1; k<=1; ++k) {

                n.k = p.k + k;

                if (n.k < 0 || n.k >= voxels_z) {
                    continue;
                }

                if (!voxels[n.i][n.j][n.k].is_medial) {
                    return false;
                }

            }
        }
    }

    return true;
    */


}

float MedialAxis::AxisDistance(const vec3 & v)
{

    float dist = FLT_MAX;
    vec3 epsvec = vec3(epsilon / 2.0f, epsilon / 2.0f, epsilon / 2.0f); //measure distance at voxel's centre

    for (int i=0; i<medial_list.size(); ++i) {

        float eachdist = ((VoxelToSpace(medial_list[i]) + epsvec) - v).length2(); //note we hang onto squared distance for now!

        if (eachdist < dist) {
            dist = eachdist;
        }

    }

    return sqrtf(dist);

}


void MedialAxis::LoadAsciiBinvox(const QString & filename, OFFFile & offfile)
{

    offfile.Clear(); 

    QString header_msg = QString("File") + filename + " has an improperly formed header";

    QFile file(filename);

    if (!file.open(QIODevice::ReadOnly | QIODevice::Text)) {
        qDebug() << "File" << filename << " can't be opened for reading";
        return;
    }

    QTextStream ifs(&file);

    //1.  read header

    //#binvox
    QString eachline = ifs.readLine();

    if (QString::compare(eachline.left(7), "#binvox") != 0) {

        qDebug() << header_msg;
        file.close();
        return;

    }

    //dimensions
    QStringList eachlist = ifs.readLine().split(" ");

    if (eachlist.size() != 4) {

        qDebug() << header_msg;
        file.close();
        return;

    }

    const int dim_x = eachlist[1].toInt();
    const int dim_y = eachlist[2].toInt();
    const int dim_z = eachlist[3].toInt();

    //translation
    eachlist = ifs.readLine().split(" ");

    if (eachlist.size() != 4) {

        qDebug() << header_msg;
        file.close();
        return;

    }

    vec3 trans_vec;
    trans_vec[0] = eachlist[1].toFloat();
    trans_vec[1] = eachlist[2].toFloat();
    trans_vec[2] = eachlist[3].toFloat();

    //scale

    eachlist = ifs.readLine().split(" ");

    if (eachlist.size() != 2) {

        qDebug() << header_msg;
        file.close();
        return;

    }

    float scale = eachlist[1].toFloat();

    //"data"

    eachline = ifs.readLine();

    if (QString::compare(eachline, "data") != 0) {

        qDebug() << header_msg;
        file.close();
        return;

    }

    //precedence for binvox format: y (fastest), z, x
    qDebug() << "Loading voxel data from " << filename << "...";
    for (int x=0; x<dim_x; ++x) {

        for (int z=0; z<dim_z; ++z) {

            eachlist = ifs.readLine().split(" ");
            for (int i=0; i<eachlist.size(); ++i) {
                if (eachlist[i].length() < 1) {
                    eachlist.removeAt(i);
                    --i;
                }
            }

            if (eachlist.size() != dim_y) {

                qDebug() << "Dimensions of data in" << filename << " does not match header's description: dim_y" << dim_y << "and size of data read" << eachlist.size();
                qDebug() << eachlist;
                file.close();
                return;

            }

            //make a square for every voxel in this column in the off file
            for (int y=0; y<dim_y; ++y) {

                if (QString::compare(eachlist[y], "1") == 0) { //put a voxel cube where there's a "1" in this file

                    //qDebug() << "putting cube at" << x << y << z;

                    vec3 v = trans_vec + vec3(x * scale / float(dim_x), y * scale / float(dim_y), z * scale / float(dim_z));
                    offfile.AddCube(v, scale / float(dim_x));

                }

            }

        }


    }

    file.close();

    qDebug() << "Voxel data successfully loaded into OFF from" << filename;

}


void MedialAxis::DoPalagyi()
{

    QVector <QVector <QVector <bool> > > v;

    //make this one padded to account for template sizes (up to 2 units out from centre point)
    v.resize(voxels.size() + 4);
    for (int i=0; i<v.size(); ++i) {
        v[i].resize(voxels[0].size() + 4);
        for (int j=0; j<v[i].size(); ++j) {
            v[i][j].resize(voxels[0][0].size() + 4);
            for (int k=0; k<v[i][j].size(); ++k) {
                v[i][j][k] = false;
            }
        }
    }

    for (int i=0; i<voxels.size(); ++i) {
        for (int j=0; j<voxels[i].size(); ++j) {
            for (int k=0; k<voxels[i][j].size(); ++k) {

                v[i+2][j+2][k+2] = voxels[i][j][k].black;

            }
        }
    }

    //order: UD, NS, EW

    int iters = 0;
    int num_deleted;

    do {

        num_deleted = 0;

        num_deleted += Deletion_From_Direction(v, 0);
        num_deleted += Deletion_From_Direction(v, 3);

        num_deleted += Deletion_From_Direction(v, 1);
        num_deleted += Deletion_From_Direction(v, 4);

        num_deleted += Deletion_From_Direction(v, 2);                
        num_deleted += Deletion_From_Direction(v, 5);

        qDebug() << "Deletions on iteration" << ++iters << ":" << num_deleted;

    } while (num_deleted > 0);


    //now update original voxels again
    for (int i=0; i<voxels.size(); ++i) {
        for (int j=0; j<voxels[i].size(); ++j) {
            for (int k=0; k<voxels[i][j].size(); ++k) {

                voxels[i][j][k].black = v[i+2][j+2][k+2];

            }
        }
    }

}

int MedialAxis::Deletion_From_Direction(QVector <QVector <QVector <bool> > > & v, int direction)
{

    QList <PalagyiTemplate> templates;

    PalagyiTemplate t1;
    t1.AddRule(-1, 1, -1, false);
    t1.AddRule(0, 1, -1, false);
    t1.AddRule(1, 1, -1, false);
    t1.AddRule(-1, 1, 0, false);
    t1.AddRule(0, 1, 0, false);
    t1.AddRule(1, 1, 0, false);
    t1.AddRule(-1, 1, 1, false);
    t1.AddRule(0, 1, 1, false);
    t1.AddRule(1, 1, 1, false);
    t1.AddRule(0, -1, 0, true);
    t1.AddRule(0, -2, 0, true);
    t1.x_condition = true;
    t1.z_condition = true;

    PalagyiTemplate t2;
    t2.AddRule(-1, 1, 0, false);
    t2.AddRule(0, 1, 0, false);
    t2.AddRule(1, 1, 0, false);
    t2.AddRule(-1, 1, -1, false);
    t2.AddRule(0, 1, -1, false);
    t2.AddRule(1, 1, -1, false);
    t2.AddRule(0, 0, 1, true);
    t2.AddRule(0, -1, 1, true);
    t2.AddRule(0, -1, 0, true);
    t2.AddRule(0, -2, 0, true);
    t2.x_condition = true;

    PalagyiTemplate t3;
    t3.AddRule(-1, 1, 0, false);
    t3.AddRule(0, 1, 0, false);
    t3.AddRule(-1, 1, -1, false);
    t3.AddRule(0, 1, -1, false);
    t3.AddRule(0, 0, 1, true);
    t3.AddRule(0, -1, 1, true);
    t3.AddRule(0, -1, 0, true);
    t3.AddRule(0, -2, 0, true);
    t3.AddRule(1, 0, 0, true);
    t3.AddRule(1, -1, 0, true);

    PalagyiTemplate t4;
    t4.AddRule(-1, 1, 1, false);
    t4.AddRule(0, 1, 1, false);
    t4.AddRule(1, 1, 1, true);
    t4.AddRule(-1, 1, 0, false);
    t4.AddRule(0, 1, 0, false);
    t4.AddRule(1, 1, 0, false);
    t4.AddRule(-1, 1, -1, false);
    t4.AddRule(0, 1, -1, false);
    t4.AddRule(1, 1, -1, false);
    t4.AddRule(1, 0, 1, true);
    t4.AddRule(1, -1, 1, true);
    t4.AddRule(0, -1, 0, true);
    t4.AddRule(0, -2, 0, true);
    t4.x_condition = true;
    t4.z_condition = true;

    PalagyiTemplate t5;
    t5.AddRule(-1, -1, -1, false);
    t5.AddRule(0, -1, -1, false);
    t5.AddRule(1, -1, -1, false);
    t5.AddRule(-1, -1, 0, false);
    t5.AddRule(0, -1, 0, false);
    t5.AddRule(1, -1, 0, false);
    t5.AddRule(-1, -1, 1, false);
    t5.AddRule(0, -1, 1, false);
    t5.AddRule(1, -1, 1, false);
    t5.AddRule(0, 1, 0, true);
    t5.x_condition = true;
    t5.z_condition = true;

    PalagyiTemplate t6;
    t6.AddRule(-1, -1, -1, false);
    t6.AddRule(0, -1, -1, false);
    t6.AddRule(1, -1, -1, false);
    t6.AddRule(-1, -1, 0, false);
    t6.AddRule(0, -1, 0, false);
    t6.AddRule(1, -1, 0, false);
    t6.AddRule(0, 1, 0, true);
    t6.AddRule(0, 0, 1, true);
    t6.x_condition = true;

    PalagyiTemplate t7;
    t7.AddRule(-1, -1, -1, false);
    t7.AddRule(0, -1, -1, false);
    t7.AddRule(-1, -1, 0, false);
    t7.AddRule(0, -1, 0, false);
    t7.AddRule(1, 0, 0, true);
    t7.AddRule(0, 1, 0, true);
    t7.AddRule(0, 0, 1, true);

    PalagyiTemplate t8;
    t8.AddRule(-1, -1, 1, false);
    t8.AddRule(0, -1, 1, false);
    t8.AddRule(1, -1, 1, true);
    t8.AddRule(-1, -1, 0, false);
    t8.AddRule(0, -1, 0, false);
    t8.AddRule(1, -1, 0, false);
    t8.AddRule(-1, -1, -1, false);
    t8.AddRule(0, -1, -1, false);
    t8.AddRule(1, -1, -1, false);
    t8.AddRule(0, 1, 0, true);
    t8.AddRule(1, 0, 1, true);
    t8.AddRule(1, 1, 1, true);
    t8.AddRule(1, -1, 1, true);
    t8.x_condition = true;
    t8.z_condition = true;

    //direction 0 - UD goes dir Y
    //direction 1 - NS goes dir Z
    //direction 2 - WE goes dir X

    if (direction == 1) {
        t1 = t1.RotateX90();
        t2 = t2.RotateX90();
        t3 = t3.RotateX90();
        t4 = t4.RotateX90();
        t5 = t5.RotateX90();
        t6 = t6.RotateX90();
        t7 = t7.RotateX90();
        t8 = t8.RotateX90();

    }
    else if (direction == 2) {
        t1 = t1.RotateZ90();
        t2 = t2.RotateZ90();
        t3 = t3.RotateZ90();
        t4 = t4.RotateZ90();
        t5 = t5.RotateZ90();
        t6 = t6.RotateZ90();
        t7 = t7.RotateZ90();
        t8 = t8.RotateZ90();
    }
    else if (direction == 3) {
        t1 = t1.RotateX180();
        t2 = t2.RotateX180();
        t3 = t3.RotateX180();
        t4 = t4.RotateX180();
        t5 = t5.RotateX180();
        t6 = t6.RotateX180();
        t7 = t7.RotateX180();
        t8 = t8.RotateX180();
    }
    else if (direction == 4) {
        t1 = t1.RotateX270();
        t2 = t2.RotateX270();
        t3 = t3.RotateX270();
        t4 = t4.RotateX270();
        t5 = t5.RotateX270();
        t6 = t6.RotateX270();
        t7 = t7.RotateX270();
        t8 = t8.RotateX270();
    }
    else if (direction == 5) {
        t1 = t1.RotateZ270();
        t2 = t2.RotateZ270();
        t3 = t3.RotateZ270();
        t4 = t4.RotateZ270();
        t5 = t5.RotateZ270();
        t6 = t6.RotateZ270();
        t7 = t7.RotateZ270();
        t8 = t8.RotateZ270();
    }

    templates.push_back(t1);
    templates.push_back(t2);
    templates.push_back(t3);
    templates.push_back(t4);
    templates.push_back(t5);
    templates.push_back(t6);
    templates.push_back(t7);
    templates.push_back(t8);

    for (int i=0; i<8; ++i) {

        if (direction == 0 || direction == 3) {
            templates.push_back(templates[i].RotateY90());
            templates.push_back(templates[i].RotateY180());
            templates.push_back(templates[i].RotateY270());
        }
        else if (direction == 1 || direction == 4) {
            templates.push_back(templates[i].RotateZ90());
            templates.push_back(templates[i].RotateZ180());
            templates.push_back(templates[i].RotateZ270());
        }
        else if (direction == 2 || direction == 5) {
            templates.push_back(templates[i].RotateX90());
            templates.push_back(templates[i].RotateX180());
            templates.push_back(templates[i].RotateX270());
        }

    }

    QList <VoxelPos> to_delete;

    for (int i=2; i<v.size()-2; ++i) {
        for (int j=2; j<v[i].size()-2; ++j) {
            for (int k=2; k<v[i][j].size()-2; ++k) {

                if (!v[i][j][k]) {
                    continue;
                }

                //test each pattern
                bool delete_it = false;
                for (int t=0; t<templates.size(); ++t) {

                    if (templates[t].Matches(v, i, j, k)) {
                        delete_it = true;
                        break;
                    }

                }

                if (delete_it) {
                    to_delete.push_back(VoxelPos(i, j, k));
                }

            }
        }
    }

    for (int i=0; i<to_delete.size(); ++i) {
        VoxelPos & p = to_delete[i];
        v[p.i][p.j][p.k] = false;
    }

    return to_delete.size();
}

void MedialAxis::DoDistanceMethod()
{

    //for all voxels compute an approximate distance to nearest outside voxel
    for (int i=0; i<voxels.size(); ++i) {

        if (i % 10 == 0) {
            qDebug() << "MedialAxis::DoDistanceMethod(): Computing for slice" << i;
        }

        for (int j=0; j<voxels[i].size(); ++j) {
            for (int k=0; k<voxels[i][j].size(); ++k) {

                //only for interior/surface voxels
                if (voxels[i][j][k].is_outside) {
                    continue;
                }

                //compute distance to nearest outside voxel, we incrementally search each outside ring
                int ring = 0;
                while (true) {

                    ++ring;

                    ClosestOutsideVoxelInRing(i, j, k, ring, voxels[i][j][k].dist, voxels[i][j][k].dist_int);

                    if (voxels[i][j][k].dist_int < INT_MAX) {
                        break;
                    }

                }

                const int fullring = int(float(ring) * 1.6f);
                for ( ; ring<fullring; ++ring) {

                    ClosestOutsideVoxelInRing(i, j, k, ring, voxels[i][j][k].dist, voxels[i][j][k].dist_int);

                }

            }
        }
    }

    //medial axes voxels are those whose distance >= their 6-neighbours
    medial_list.clear();

    for (int i=1; i<voxels.size()-1; ++i) {
        for (int j=1; j<voxels[i].size()-1; ++j) {
            for (int k=1; k<voxels[i][j].size()-1; ++k) {

                //only for interior/surface voxels
                if (voxels[i][j][k].is_outside || voxels[i][j][k].is_surface) {
                    voxels[i][j][k].black = false;
                    continue;
                }

                //check with 26-neighbours
                voxels[i][j][k].black = true;
                for (int ii=i-1; ii<=i+1; ++ii) {
                    for (int jj=j-1; jj<=j+1; ++jj) {
                        for (int kk=k-1; kk<=k+1; ++kk ) {

                            //only for interior/surface voxels
                            if (voxels[ii][jj][kk].is_outside) {
                                continue;
                            }

                            //if one of my 26-neighbours has a greater dist than i do, then i'm not a medial axis voxel
                            if (voxels[ii][jj][kk].dist_int > voxels[i][j][k].dist_int) {
                                voxels[i][j][k].black = false;
                            }

                        }
                    }
                }

            }
        }
    }

}

void MedialAxis::ClosestOutsideVoxelInRing(const int i, const int j, const int k, const int ring, float & dist, int & dist_int)
{

    for (int ii=qMax(i-ring,0); ii<=qMin(i+ring, voxels_x-1); ++ii) {
        for (int jj=qMax(j-ring,0); jj<=qMin(j+ring, voxels_y-1); ++jj) {
            for (int kk=qMax(k-ring,0); kk<=qMin(k+ring, voxels_z-1); ++kk) {

                if (abs(ii - i) != ring || abs(jj - j) != ring || abs(kk - k) != ring) {
                    continue;
                }

                if (voxels[ii][jj][kk].is_outside) {

                    const float eachdist = (vec3(i, j, k) - vec3(ii, jj, kk)).length();
                    dist = qMin(dist, eachdist);

                    //const int eachdist_int = abs(ii - i) + abs(jj - j) + abs(kk - k);
                    const int eachdist_int = qMax(qMax(abs(ii - i), abs(jj - j)), abs(kk - k));
                    dist_int = qMin(dist_int, eachdist_int);

                }

            }
        }
    }

}

void MedialAxis::DoSweepy() {

    int sweep_dir = -1;

    while (true) {

        //change sweeping direction
        ++sweep_dir;
        sweep_dir = sweep_dir % 6;

        int n_removed = 0;

        QVector <QVector <QVector <Voxel> > > new_voxels = voxels;

        for (int i=1; i<voxels_x-1; ++i) {
            for (int j=1; j<voxels_y-1; ++j) {
                for (int k=1; k<voxels_z-1; ++k) {

                    //only for interior/surface voxels
                    if (!voxels[i][j][k].black) {
                        continue;
                    }

                    if (DoSweepy_VoxThin(i, j, k)) {
                        continue;
                    }

                    switch (sweep_dir) {

                    case 0: //-x

                        if (!voxels[i+1][j][k].black && voxels[i-1][j][k].black) {
                            new_voxels[i][j][k].black = false; ++n_removed;
                        }

                        break;

                    case 1: //x

                        if (!voxels[i-1][j][k].black && voxels[i+1][j][k].black) {
                            new_voxels[i][j][k].black = false; ++n_removed;
                        }

                        break;

                    case 2: //-y

                        if (!voxels[i][j+1][k].black && voxels[i][j-1][k].black) {
                            new_voxels[i][j][k].black = false; ++n_removed;
                        }

                        break;

                    case 3: //y

                        if (!voxels[i][j-1][k].black && voxels[i][j+1][k].black) {
                            new_voxels[i][j][k].black = false; ++n_removed;
                        }

                        break;

                    case 4: //-z

                        if (!voxels[i][j][k+1].black && voxels[i][j][k-1].black) {
                            new_voxels[i][j][k].black = false; ++n_removed;
                        }

                        break;

                    case 5: //z

                        if (!voxels[i][j][k-1].black && voxels[i][j][k+1].black) {
                            new_voxels[i][j][k].black = false; ++n_removed;
                        }

                        break;

                    }


                }
            }
        }

        voxels = new_voxels;

        qDebug() << "Voxels removed this pass:" << n_removed;

        if (n_removed == 0) {
            break;
        }

    }
}

bool MedialAxis::DoSweepy_VoxThin(const int i, const int j, const int k) {

    return (!voxels[i-1][j][k].black && !voxels[i+1][j][k].black) ||
            (!voxels[i][j-1][k].black && !voxels[i][j+1][k].black) ||
            (!voxels[i][j][k-1].black && !voxels[i][j][k+1].black);

}
