#ifndef MEDIALAXIS_H
#define MEDIALAXIS_H

#include <QtCore>
#include <QtOpenGL>

#include <cfloat>

#include "algebra3.h"
#include "OFFFile.h"

struct Voxel
{

    Voxel() :
        black(false),
        is_surface(false),
        is_outside(false),
        dist(FLT_MAX),
        dist_int(INT_MAX)
    { }

    bool black;
    bool is_surface;
    bool is_outside;

    float dist;
    int dist_int;

};

struct VoxelPos
{

    VoxelPos() :
        i(0),
        j(0),
        k(0)
    { }

    VoxelPos(const int _i, const int _j, const int _k) :
        i(_i),
        j(_j),
        k(_k)
    { }

    bool operator==(const VoxelPos & o) const { return i == o.i && j == o.j && k == o.k; }

    int i;
    int j;
    int k;

};

struct PalagyiRule
{

    PalagyiRule(int x, int y, int z, int b) :
        i(x),
        j(y),
        k(z),
        black(b)
    {

    }

    int i;
    int j;
    int k;
    bool black;

};

struct PalagyiTemplate
{

    PalagyiTemplate() :
        x_condition(false),
        y_condition(false),
        z_condition(false)
    {

    }

    void AddRule(int x, int y, int z, bool b) {
        rules.push_back(PalagyiRule(x, y, z, b));
    }

    PalagyiTemplate RotateX90() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(rules[i].i, -rules[i].k, rules[i].j, rules[i].black);
        }

        t.x_condition = x_condition;
        t.y_condition = z_condition;
        t.z_condition = y_condition;

        return t;
    }

    PalagyiTemplate RotateX180() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(rules[i].i, -rules[i].j, -rules[i].k, rules[i].black);
        }

        t.x_condition = x_condition;
        t.y_condition = y_condition;
        t.z_condition = z_condition;

        return t;
    }

    PalagyiTemplate RotateX270() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(rules[i].i, rules[i].k, -rules[i].j, rules[i].black);
        }

        t.x_condition = x_condition;
        t.y_condition = z_condition;
        t.z_condition = y_condition;

        return t;
    }

    PalagyiTemplate RotateY90() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(-rules[i].k, rules[i].j, rules[i].i, rules[i].black);
        }

        t.x_condition = z_condition;
        t.y_condition = y_condition;
        t.z_condition = x_condition;

        return t;
    }

    PalagyiTemplate RotateY180() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(-rules[i].i, rules[i].j, -rules[i].k, rules[i].black);
        }

        t.x_condition = x_condition;
        t.y_condition = y_condition;
        t.z_condition = z_condition;

        return t;
    }

    PalagyiTemplate RotateY270() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(rules[i].k, rules[i].j, -rules[i].i, rules[i].black);
        }

        t.x_condition = z_condition;
        t.y_condition = y_condition;
        t.z_condition = x_condition;

        return t;
    }

    PalagyiTemplate RotateZ90() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(-rules[i].j, rules[i].i, rules[i].k, rules[i].black);
        }

        t.x_condition = y_condition;
        t.y_condition = x_condition;
        t.z_condition = z_condition;

        return t;
    }

    PalagyiTemplate RotateZ180() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(-rules[i].i, -rules[i].j, rules[i].k, rules[i].black);
        }

        t.x_condition = x_condition;
        t.y_condition = y_condition;
        t.z_condition = z_condition;

        return t;
    }

    PalagyiTemplate RotateZ270() {

        PalagyiTemplate t;

        for (int i=0; i<rules.size(); ++i) {
            t.AddRule(rules[i].j, -rules[i].i, rules[i].k, rules[i].black);
        }

        t.x_condition = y_condition;
        t.y_condition = x_condition;
        t.z_condition = z_condition;

        return t;
    }

    bool Matches(QVector <QVector <QVector <bool> > > & v, int x, int y, int z) {

        if (x_condition && !v[x-1][y][z] && !v[x+1][y][z]) {
            return false;
        }

        if (y_condition && !v[x][y-1][z] && !v[x][y+1][z]) {
            return false;
        }

        if (z_condition && !v[x][y][z-1] && !v[x][y][z+1]) {
            return false;
        }

        for (int i=0; i<rules.size(); ++i) {

            if (v[x+rules[i].i][y+rules[i].j][z+rules[i].k] != rules[i].black) {
                return false;
            }

        }

        return true;

    }

    bool x_condition;
    bool y_condition;
    bool z_condition;

    QList <PalagyiRule> rules;

};

class MedialAxis
{

public:

    MedialAxis();   

    //my own implementation
    vec3 VoxelToSpace(const VoxelPos & p);
    VoxelPos SpaceToVoxel(const vec3 & v);   

    void CreateFromSurface(const QVector <vec3> & verts, const QVector <QVector <int> > & faces, const float eps);

    void WriteSurface(const QString & filename);
    void WriteMedialAxis(const QString & filename);

    float AxisDistance(const vec3 & v);

    void DrawGL();   

    //binvox/thinvox-dependent implementation... this doesn't work out since it doesn't do medial surfaces
    void LoadAsciiBinvox(const QString & filename, OFFFile & offfile);

    //this 3d thinning algorithm works well enough (Palagyi '02, 3-subiteration 3d thinning algo)
    void DoPalagyi();
    int Deletion_From_Direction(QVector <QVector <QVector <bool> > > & v, int direction);

    void DoDistanceMethod();
    void ClosestOutsideVoxelInRing(const int i, const int j, const int k, const int ring, float & dist, int & dist_int);

    void DoSweepy();
    bool DoSweepy_VoxThin(const int i, const int j, const int k);

    //helpful routine for other stuff
    QList <VoxelPos> GetSurfaceList(const QVector <vec3> & verts, const QVector <QVector <int> > & faces, const float eps);
    QList <vec3> SurfaceListToVec(const QList <VoxelPos> & s_list);

private:   

    bool IsVoxelInsideMedial(const VoxelPos & p);

    vec3 bbox_min;
    vec3 bbox_max;

    int voxels_x;
    int voxels_y;
    int voxels_z;

    float epsilon;

    QVector <QVector <QVector <Voxel> > > voxels;

    QList <VoxelPos> surface_list;
    QList <VoxelPos> medial_list;

    OFFFile surface_geom;
    OFFFile medial_geom;

};

#endif // MEDIALAXIS_H
