#include "orbitcamera.h"

OrbitCamera::OrbitCamera() :
    neardist(0.1),
    fardist(4.0),
    cam_width(2.0f),
    theta(45.0f),
    phi(45.0f),
    viewdist(1.0f),
    lookat(0, 0, 0),
    interp_active(false),
    interp_cam_start(0),
    interp_cam_end(0)
{
}

void OrbitCamera::Update()
{   

    if (interp_active) {

        if (interp_cam_end->theta - interp_cam_start->theta > 180.0f) {
            interp_cam_end->theta -= 360.0f;
        }
        else if (interp_cam_end->theta - interp_cam_start->theta < -180.0f) {
            interp_cam_end->theta += 360.0f;
        }

        float interp_time = InterpTime();
        theta = InterpValue_Cosine(interp_cam_start->theta, interp_cam_end->theta, interp_time);
        phi = InterpValue_Cosine(interp_cam_start->phi, interp_cam_end->phi, interp_time);
        viewdist = InterpValue_Cosine(interp_cam_start->viewdist, interp_cam_end->viewdist, interp_time);
        lookat = InterpValue_Cosine(interp_cam_start->lookat, interp_cam_end->lookat, interp_time);
        cam_width = InterpValue_Cosine(interp_cam_start->cam_width, interp_cam_end->cam_width, interp_time);

        if (InterpTime() >= 1.0f) {
            interp_active = false;
        }

    }

    eye = SphericalToCartesian(theta, phi, viewdist) + lookat;
    up = SphericalToCartesian(theta, phi - 90.0f, 1.0f);

}

void OrbitCamera::DrawGL() const
{

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    gluPerspective(75.0, float(win_width) / float(win_height), neardist, fardist);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(eye[0], eye[1], eye[2], lookat[0], lookat[1], lookat[2], up[0], up[1], up[2]);

}

void OrbitCamera::DrawGL_Ortho() const
{

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0.0f, win_width, 0.0f, win_height, -1.0f, 1.0f);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

}

void OrbitCamera::DrawGL_3DOrtho() const
{

    float aspectr = float(win_height) / float(win_width);

    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    //glOrtho(-cam_width/2.0, cam_width/2.0, -cam_width/2.0 * aspectr, cam_width/2.0 * aspectr, -fardist, fardist);
    glOrtho(-cam_width/2.0, cam_width/2.0, -cam_width/2.0 * aspectr, cam_width/2.0 * aspectr, neardist, fardist);

    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    gluLookAt(eye[0], eye[1], eye[2], lookat[0], lookat[1], lookat[2], up[0], up[1], up[2]);

}

void OrbitCamera::Orbit(float dx, float dy)
{

    theta += dx * 0.25f;
    phi += dy * 0.25f;

    theta += (theta > 180.0f)? -360.0f : 0.0f;
    theta += (theta < -180.0f)? 360.0f : 0.0f;

    //phi = qMax(phi, 0.0f);
    //phi = qMin(phi, 180.0f);

}

void OrbitCamera::Pan(float dx, float dy)
{

    vec3 up_vec = up;
    vec3 forward_vec = lookat - eye;
    forward_vec.normalize();

    vec3 right_vec = forward_vec ^ up_vec;

    float move_factor = viewdist * 0.0002;

    vec3 displacement = right_vec * dx + up_vec * dy;

    lookat -= displacement * move_factor;

}

vec3 OrbitCamera::SphericalToCartesian(float t, float p, float r) const
{

    float theta_rad = t * DEG_TO_RAD;
    float phi_rad = p * DEG_TO_RAD;

    vec3 v;

    v[0] = r * cosf(theta_rad) * sinf(phi_rad);
    v[1] = r * cosf(phi_rad);
    v[2] = r * sinf(theta_rad) * sinf(phi_rad);


    return v;

}

vec3 OrbitCamera::Up() const
{

    return up;

}


vec3 OrbitCamera::Eye() const
{

    return eye;

}

vec3 OrbitCamera::LookAt() const
{

    return lookat;

}

void OrbitCamera::SetLookAt(const vec3 & p)
{
    lookat = p;
}

float OrbitCamera::Theta() const
{
    return theta;
}

void OrbitCamera::SetTheta(float f)
{
    theta = f;
}

float OrbitCamera::Phi() const
{
    return phi;
}

void OrbitCamera::SetPhi(float f)
{
    phi = f;
}

vec3 OrbitCamera::ViewDir() const
{
    return (lookat - Eye()).normalize();
}

float OrbitCamera::ViewDist() const
{
    return viewdist;
}

void OrbitCamera::SetViewDist(float f)
{
    viewdist = f;
}

void OrbitCamera::SetWinWidth(int i)
{
    win_width = i;
}

void OrbitCamera::SetWinHeight(int i)
{
    win_height = i;
}

void OrbitCamera::SetCamWidth(float f)
{
    cam_width = f;
}

float OrbitCamera::CamWidth() const
{
    return cam_width;
}

float OrbitCamera::InterpTime() const
{

    float interp = float(interp_time.elapsed()) / interp_duration;

    interp = qMin(1.0f, interp);
    interp = qMax(0.0f, interp);

    return interp;

}

template <typename T>
T OrbitCamera::InterpValue_Linear(T v1, T v2, float interp) const
{

    return v1 * (1.0f - interp) + v2 * interp;

}

template <typename T>
T OrbitCamera::InterpValue_Cosine(T v1, T v2, float interp) const
{

    float cos_interp = 1.0f - (cosf(interp * ALG_M_PI) + 1.0f) / 2.0f;
    return v1 * (1.0f - cos_interp) + v2 * cos_interp;

}

void OrbitCamera::SetInterpCamera(OrbitCamera & cam, float duration)
{

    interp_time.start();
    interp_duration = duration;
    interp_active = true;

    if (!interp_cam_start) {
        interp_cam_start = new OrbitCamera();
    }
    *interp_cam_start = *this;

    if (!interp_cam_end) {
        interp_cam_end = new OrbitCamera();
    }
    *interp_cam_end = cam;

}

void OrbitCamera::ZoomIn()
{
    viewdist *= 0.8f;
    cam_width *= 0.8f;
}

void OrbitCamera::ZoomOut()
{
    viewdist *= 1.25f;
    cam_width *= 1.25f;
}
