#include "plane.h"

Plane::Plane()
{

    bbox_min_t = 0.0f;
    bbox_min_b = 0.0f;
    bbox_max_t = 0.0f;
    bbox_max_b = 0.0f;

}

void Plane::SetPts(const vec3 & p1, const vec3 & p2, const vec3 & p3)
{

    vec3 b1 = p2 - p1;
    b1.normalize();
    vec3 b2 = p3 - p1;
    b2.normalize();

    vec3 norm = b1 ^ b2;
    b2 = b1 ^ norm;
    b2.normalize();

    SetTNB(b1, norm, b2);

    d = N() * p1;
}

void Plane::SetD(const float val)
{
    d = val;
}

float Plane::D() const
{
    return d;
}

vec3 Plane::GetCentre() const
{
    return n * d;
}

void Plane::DrawLoopGL(const vec3 & centre, const float t_size, const float b_size) const
{
    glBegin(GL_LINE_LOOP);
    glVertex3dv(&(centre + t * t_size + b * b_size)[0]);
    glVertex3dv(&(centre + t * t_size - b * b_size)[0]);
    glVertex3dv(&(centre - t * t_size - b * b_size)[0]);
    glVertex3dv(&(centre - t * t_size + b * b_size)[0]);
    glEnd();
}

void Plane::DrawQuadGL(const vec3 & centre, const float t_size, const float b_size) const
{
    glBegin(GL_QUADS);
    glVertex3dv(&(centre + t * t_size + b * b_size)[0]);
    glVertex3dv(&(centre + t * t_size - b * b_size)[0]);
    glVertex3dv(&(centre - t * t_size - b * b_size)[0]);
    glVertex3dv(&(centre - t * t_size + b * b_size)[0]);
    glEnd();
}

void Plane::DrawGL() const
{

    glBegin(GL_LINE_LOOP);
    glVertex3dv(&(n * d + t + b)[0]);
    glVertex3dv(&(n * d + t - b)[0]);
    glVertex3dv(&(n * d - t - b)[0]);
    glVertex3dv(&(n * d - t + b)[0]);
    glEnd();

    glBegin(GL_LINES);
    glVertex3dv(&(n * d + t)[0]);
    glVertex3dv(&(n * d - t)[0]);
    glVertex3dv(&(n * d + b)[0]);
    glVertex3dv(&(n * d - b)[0]);
    glEnd();

}

void Plane::DrawBoundingBoxGL() const
{

    glBegin(GL_LINE_LOOP);
    glVertex3dv(&(n * d + t * bbox_max_t + b * bbox_max_b)[0]);
    glVertex3dv(&(n * d + t * bbox_max_t + b * bbox_min_b)[0]);
    glVertex3dv(&(n * d + t * bbox_min_t + b * bbox_min_b)[0]);
    glVertex3dv(&(n * d + t * bbox_min_t + b * bbox_max_b)[0]);
    glEnd();

}

float Plane::PointDistance(const vec3 & p) const
{

    vec3 w = p - (n * d);
    return fabs(n * w);

}

bool Plane::RayIntersect(const vec3 & p0, const vec3 & dir, vec3 & intersect) const
{

    float denom = n * dir;

    if (fabs(denom) < 0.0001f) {
        return false;
    }

    float u = (n * (n * d - p0)) / denom;
    intersect = p0 + dir * u;

    return true;

}

bool Plane::LineSegmentIntersect(const vec3 & p0, const vec3 & p1, vec3 & intersect) const
{

    float u;
    return LineSegmentIntersect(p0, p1, intersect, u);

}

bool Plane::LineSegmentIntersect(const vec3 & p0, const vec3 & p1, vec3 & intersect, float & u) const
{

    vec3 dir = p1 - p0;

    float denom = n * dir;

    if (fabs(denom) < 0.0001f) {
        return false;
    }

    u = (n * (n * d - p0)) / denom;
    intersect = p0 + dir * u;

    return (u >= 0.0f && u <= 1.0f);

}

void Plane::BestFitPlane(const QList <vec3> & points, vec3 & eigen1, vec3 & eigen2, vec3 & eigen3, vec3 & centroid)
{

    //compute centroid
    vec3 kOrigin(0,0,0);
    double recip = 1.0 / double(points.size());

    for (int i=0; i<points.size(); i++)
    {
        kOrigin += points[i];
    }

    kOrigin *= recip;

    //compute covariance matrix

    double fSumXX=0;
    double fSumXY=0;
    double fSumXZ=0;

    double fSumYY=0;
    double fSumYZ=0;
    double fSumZZ=0;

    for (int i=0; i<points.size(); i++)
    {

        vec3 kDiff = points[i] - kOrigin;

        fSumXX += kDiff[0] * kDiff[0]; // sum of the squares of differences
        fSumXY += kDiff[0] * kDiff[1]; // sum of the squares of differences
        fSumXZ += kDiff[0] * kDiff[2]; // sum of the squares of differences

        fSumYY += kDiff[1] * kDiff[1];
        fSumYZ += kDiff[1] * kDiff[2];
        fSumZZ += kDiff[2] * kDiff[2];

    }

    fSumXX *= recip;
    fSumXY *= recip;
    fSumXZ *= recip;
    fSumYY *= recip;
    fSumYZ *= recip;
    fSumZZ *= recip;

    // setup the eigensolver
    Eigen kES;

    kES.mElement[0][0] = fSumXX;
    kES.mElement[0][1] = fSumXY;
    kES.mElement[0][2] = fSumXZ;

    kES.mElement[1][0] = fSumXY;
    kES.mElement[1][1] = fSumYY;
    kES.mElement[1][2] = fSumYZ;

    kES.mElement[2][0] = fSumXZ;
    kES.mElement[2][1] = fSumYZ;
    kES.mElement[2][2] = fSumZZ;

    // compute eigenstuff, smallest eigenvalue is in last position
    kES.DecrSortEigenStuff();

    //normal direction will be in direction of least significant eigenvector
    //and plane will cross the centroid of the points
    vec3 kNormal;        

    kNormal[0] = kES.mElement[0][2];
    kNormal[1] = kES.mElement[1][2];
    kNormal[2] = kES.mElement[2][2];

    kNormal.normalize();

    SetN(kNormal);
    SetD(kNormal * kOrigin);

    eigen1 = vec3(kES.mElement[0][0], kES.mElement[1][0], kES.mElement[2][0]);
    eigen2 = vec3(kES.mElement[0][1], kES.mElement[1][1], kES.mElement[2][1]);
    eigen3 = vec3(kES.mElement[0][2], kES.mElement[1][2], kES.mElement[2][2]);
    centroid = kOrigin;

}

vec2 Plane::GetPoint2D(const vec3 & p) const
{

    vec3 p_0 = p - (n * d);
    return vec2(p_0 * T(), p_0 * B());

}

vec3 Plane::GetPoint3D(const vec2 & p) const
{

    return (T() * p[0]) + (B() * p[1]) + (n * d);

}

vec3 Plane::MirrorPoint(const vec3 & p) const
{         

    //vec3 newpos = p - n * (p * n - d) * 2.0f;

    //qDebug() << "Flipping p" << p << "of dist" << p * n - d << "to" << newpos << "of dist" << newpos * n - d << "d" << d;

    vec3 rel_p = p - (n * d);
    vec3 n_part = n * (n * rel_p);
    vec3 newpos = p - (n_part * 2.0f);

    return newpos;

}

void Plane::SetBoundingBox(const float min_t, const float min_b, const float max_t, const float max_b)
{

    bbox_min_t = min_t;
    bbox_min_b = min_b;
    bbox_max_t = max_t;
    bbox_max_b = max_b;

}
